/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <gtk/gtkwidget.h>
#include <gtk/gtkalignment.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkcombo.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtkitemfactory.h>
#include <gtk/gtkmenubar.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktable.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkstock.h>
#include <gtk/gtktextview.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkcheckmenuitem.h>
#include <gtk/gtkmenu.h>

#include "intl.h"
#include "x_alloc.h"
#include "logging.h"

#include "ui.h"

#include "pixmaps/xetpan.xpm"

#include "gtk-utils.h"
#include "gtk-messageview.h"
#include "gtk-toolbar.h"
#include "identity.h"
#include "utils.h"
#include "mta.h"

#include "gtk-composewindow.h"


typedef struct XePGTKComposeWindowHeader_s XePGTKComposeWindowHeader;

struct XePGTKComposeWindowHeader_s
{
  XePGTKComposeWindow *uicompose;
  int index;
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *add;
  GtkWidget *remove;
};

typedef struct 
{
  char *path;
  XePUtilsCharset charset_index;
  char *type;
} XePGTKCharsetMenu;


XePGTKCharsetMenu charset_menu[] =
  {
    /* First item must be the radioitem leader */
    {N_("/Edit/Character Encoding/Unicode (UTF-8)"), XEP_CHARSET_UTF8, "<RadioItem>"}, 
    {"/Edit/Character Encoding/sep1", 0, "<Separator>"},
    {N_("/Edit/Character Encoding/_Asian"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/_Chinese Traditional"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Chinese Traditional/_Big5"), XEP_CHARSET_BIG5, NULL},
    {N_("/Edit/Character Encoding/Asian/Chinese Traditional/_EUC-TW"), XEP_CHARSET_EUC_TW, NULL},
    {N_("/Edit/Character Encoding/Asian/Chinese _Simplified"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Chinese Simplified/_ISO-2022-CN"), XEP_CHARSET_ISO_2022_CN, NULL},
    {N_("/Edit/Character Encoding/Asian/Chinese Simplified/ISO-2022-CN-_EXT"), XEP_CHARSET_ISO_2022_CN_EXT, NULL},
    {N_("/Edit/Character Encoding/Asian/Chinese Simplified/_GB2312"), XEP_CHARSET_GB2312, NULL},
    {N_("/Edit/Character Encoding/Asian/_Japanese"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Japanese/_ISO-2022-JP"), XEP_CHARSET_ISO_2022_JP, NULL},
    {N_("/Edit/Character Encoding/Asian/Japanese/ISO-2022-JP-_2"), XEP_CHARSET_ISO_2022_JP_2, NULL},
    {N_("/Edit/Character Encoding/Asian/Japanese/_EUC-JP"), XEP_CHARSET_EUC_JP, NULL},
    {N_("/Edit/Character Encoding/Asian/Japanese/Shift-_JIS"), XEP_CHARSET_SHIFT_JIS, NULL},
    {N_("/Edit/Character Encoding/Asian/_Korean"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Korean/_EUC-KR"), XEP_CHARSET_EUC_KR, NULL},
    {N_("/Edit/Character Encoding/Asian/_Thai"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Thai/_ISO-8859-11"), XEP_CHARSET_ISO_8859_11, NULL},
    {N_("/Edit/Character Encoding/Asian/Thai/_TIS-620"), XEP_CHARSET_TIS_620, NULL},
    {N_("/Edit/Character Encoding/Asian/Tu_rkish"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Turkish/_ISO-8859-9"), XEP_CHARSET_ISO_8859_9, NULL},
    {N_("/Edit/Character Encoding/Asian/Turkish/_Windows-1254"), XEP_CHARSET_WINDOWS_1254, NULL},
    {N_("/Edit/Character Encoding/Asian/_Vietnamese"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Asian/Vietnamese/_VISCII"), XEP_CHARSET_VISCII, NULL},
    {N_("/Edit/Character Encoding/Asian/Vietnamese/_Windows-1258"), XEP_CHARSET_WINDOWS_1258, NULL},
    {N_("/Edit/Character Encoding/_European"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/European/Balticz"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/European/Baltic/_ISO-8859-4"), XEP_CHARSET_ISO_8859_4, NULL},
    {N_("/Edit/Character Encoding/European/Baltic/ISO-8859-1_3"), XEP_CHARSET_ISO_8859_13, NULL},
    {N_("/Edit/Character Encoding/European/_Celtic (ISO-8859-14)"), XEP_CHARSET_ISO_8859_14, NULL},
    {N_("/Edit/Character Encoding/European/Central _Europe"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/European/Central Europe/_ISO-8859-2"), XEP_CHARSET_ISO_8859_2, NULL},
    {N_("/Edit/Character Encoding/European/Central Europe/_Windows-1250"), XEP_CHARSET_WINDOWS_1250, NULL},
    {N_("/Edit/Character Encoding/European/C_yrillic"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/European/Cyrillic/_ISO-8859-5"), XEP_CHARSET_ISO_8859_5, NULL},
    {N_("/Edit/Character Encoding/European/Cyrillic/_MacCyrillic"), XEP_CHARSET_MAC_CYRILLIC, NULL},
    {N_("/Edit/Character Encoding/European/Cyrillic/_Russian (KOI8-R)"), XEP_CHARSET_KOI8_R, NULL},
    {N_("/Edit/Character Encoding/European/Cyrillic/_Ukrainian (KOI8-U)"), XEP_CHARSET_KOI8_U, NULL},
    {N_("/Edit/Character Encoding/European/Cyrillic/_Windows-1251"), XEP_CHARSET_WINDOWS_1251, NULL},
    {N_("/Edit/Character Encoding/European/_Greek (ISO-8859-7)"), XEP_CHARSET_ISO_8859_7, NULL},
    {N_("/Edit/Character Encoding/European/_Nordic (ISO-8859-10)"), XEP_CHARSET_ISO_8859_10, NULL},
    {N_("/Edit/Character Encoding/European/Ro_manian (ISO-8859-16)"), XEP_CHARSET_ISO_8859_16, NULL},
    {N_("/Edit/Character Encoding/European/_Southern Europe (ISO-8859-3)"), XEP_CHARSET_ISO_8859_3, NULL},
    {N_("/Edit/Character Encoding/European/_Western Europe"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/European/Western Europe/_ISO-8859-15"), XEP_CHARSET_ISO_8859_15, NULL},
    {N_("/Edit/Character Encoding/European/Western Europe/ISO-8859-_1"), XEP_CHARSET_ISO_8859_1, NULL},
    {N_("/Edit/Character Encoding/_Oriental"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Oriental/_Arabic"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Oriental/Arabic/_ISO-8859-6"), XEP_CHARSET_ISO_8859_6, NULL},
    {N_("/Edit/Character Encoding/Oriental/Arabic/_Windows-1256"), XEP_CHARSET_WINDOWS_1256, NULL},
    {N_("/Edit/Character Encoding/_Oriental/_Farsi (ISIRI-3342)"), XEP_CHARSET_ISIRI_3342, NULL},
    {N_("/Edit/Character Encoding/Oriental/_Hebrew"), 0, "<Branch>"},
    {N_("/Edit/Character Encoding/Oriental/Hebrew/_ISO-8859-8 (implicit)"), XEP_CHARSET_ISO_8859_8_I, NULL},
    {N_("/Edit/Character Encoding/Oriental/Hebrew/ISO-8859-8 (_visual)"), XEP_CHARSET_ISO_8859_8, NULL},
    {N_("/Edit/Character Encoding/Oriental/Hebrew/_Windows-1255"), XEP_CHARSET_WINDOWS_1255, NULL}
  };


static GtkWidget *
GtkComposewindow_menu_create(XePGTKComposeWindow *uicompose);

static void 
GtkComposewindow_set_default_encoding(XePGTKComposeWindow *uicompose,
				      const char *charset);

static void
GtkComposewindow_populate_from_combo(XePGTKComposeWindow *uicompose);

static void
GtkComposewindow_text_view_popup_cb(GtkTextView *text_view,
				    GtkMenu *menu,
				    gpointer data);
static gboolean
GtkComposewindow_delete_cb(GtkWidget *widget,
			   GdkEvent *event,
			   gpointer data);

static void
GtkComposewindow_file_send_cb(gpointer data,
			      guint action,
			      GtkWidget *widget);

static void
GtkComposewindow_file_draft_cb(gpointer data,
			       guint action,
			       GtkWidget *widget);

static void
GtkComposewindow_file_close_cb(gpointer data,
			       guint action,
			       GtkWidget *widget);

static gboolean
GtkComposewindow_delete_cb(GtkWidget *widget,
			   GdkEvent *event,
			   gpointer data);

static void
GtkComposewindow_edit_add_header_cb(gpointer data,
				    guint action,
				    GtkWidget *widget);

static void
GtkComposewindow_edit_charset_cb(gpointer data,
				 guint action,
				 GtkWidget *widget);

static void
GtkComposewindow_add_header(XePGTKComposeWindow *uicompose,
			    int row,
			    XePMessageHeaderType header_type,
			    const char *other_header,
			    const char *header_value);

static void
GtkComposewindow_add_header_real(XePGTKComposeWindow *uicompose,
				 int row,
				 const char *header_name,
				 const char *header_value);

static void
GtkComposewindow_add_header_cb(GtkWidget *button, gpointer data);

static void
GtkComposewindow_remove_header_cb(GtkWidget *button, gpointer data);

static void
GtkComposewindow_show_source_cb(GtkMenuItem *menu_item, gpointer data);

static void
GtkComposewindow_save_source_as_cb(GtkMenuItem *menu_item, gpointer data);

static int
GtkComposewindow_save_source_as(void *amessage, const char *filename);

static void
GtkComposewindow_build_message(XePGTKComposeWindow *uicompose);

static void
GtkComposewindow_sent_cb(void *auicompose, void *aresult);

/* Function Defintions */
XePGTKComposeWindow *
a_GtkComposewindow_create(XetPan *xetpan, char *terminal)
{
  XePGTKComposeWindow *uicompose;
  XePGTKMessageView *mview;
  GtkWidget *toplevel;
  GtkWidget *topvbox;
  GtkWidget *menubar;
  GtkWidget *toolbar_hbox;
  GtkWidget *main_frame;
  GtkWidget *vpane;
  GtkWidget *scrollw;
  GtkWidget *viewport;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *combo;
  GtkWidget *mvbox;
  GtkWidget *hbox;
  GtkWidget *entry;
  GtkWidget *message_frame;
  
  uicompose = (XePGTKComposeWindow *) a_UI_composewindow_new();
  if (uicompose == NULL)
    return NULL;

  uicompose->type = XETPAN_UI_GTK;
  uicompose->xetpan = xetpan;
  uicompose->headers = NULL;

  toplevel = a_GtkUtils_toplevel_new(terminal,
				     _("Compose Mail"),
				     xetpan_xpm,
				     460, 460);
  uicompose->toplevel = toplevel;

  topvbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(topvbox);
  gtk_container_add(GTK_CONTAINER(toplevel), topvbox);

  /* Menu bar */

  menubar = GtkComposewindow_menu_create(uicompose);
  gtk_widget_show(menubar);
  gtk_box_pack_start(GTK_BOX(topvbox), menubar, FALSE, FALSE, 0);

  uicompose->menubar = menubar;

  toolbar_hbox = a_GtkToolbar_composewindow_new(uicompose);
  gtk_widget_show(toolbar_hbox);

  gtk_box_pack_start(GTK_BOX(topvbox), toolbar_hbox, FALSE, FALSE, 0);

  /* Main frame */

  main_frame = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(main_frame), GTK_SHADOW_OUT);
  gtk_widget_show(main_frame);
  gtk_box_pack_start(GTK_BOX(topvbox), main_frame, TRUE, TRUE, 0);

  /* Vertically paned */

  vpane = gtk_vpaned_new();
  gtk_widget_show(vpane);
  gtk_container_add(GTK_CONTAINER(main_frame), vpane);
  a_GtkUtils_paned_shrink_control(vpane, &(uicompose->header_pane_position));

  /* Headers view scrolled window */
  
  scrollw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollw),
				 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_ALWAYS);
  gtk_widget_show(scrollw);
  gtk_paned_pack1(GTK_PANED(vpane), scrollw, FALSE, FALSE);

  /* Headers view viewport */
  
  viewport = gtk_viewport_new(NULL, NULL);
  gtk_viewport_set_shadow_type(GTK_VIEWPORT(viewport), GTK_SHADOW_NONE);
  gtk_widget_show(viewport);
  gtk_container_add(GTK_CONTAINER(scrollw), viewport);

  /* Header view table */
  
  table = gtk_table_new(1, 4, FALSE);
  gtk_widget_show(table);
  gtk_container_add(GTK_CONTAINER(viewport), table);

  /* From label */

  label = a_GtkUtils_label_new_bold_large(_("From:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  gtk_widget_show(label);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
		   GTK_SHRINK | GTK_FILL, GTK_SHRINK, 12, 0);

  /* From: combo box */

  combo = gtk_combo_new();

  gtk_widget_show(combo);
  gtk_table_attach(GTK_TABLE(table), combo, 1, 2, 0, 1,
		   GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);


  /* Message vbox */

  mvbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(mvbox);
  gtk_paned_pack2(GTK_PANED(vpane), mvbox, TRUE, TRUE);

  /* Subject hbox */

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(mvbox), hbox, FALSE, FALSE, 0);

  /* Subject label */

  label = a_GtkUtils_label_new_bold(_("Subject:"));
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 12);

  /* Subject entry */

  entry = gtk_entry_new();
  gtk_widget_show(entry);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  /* Message frame */

#if 0
  message_frame = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(message_frame), GTK_SHADOW_NONE);
#endif
  message_frame = gtk_alignment_new(0.0, 0.0, 1.0, 1.0);
  gtk_widget_show(message_frame);
  gtk_box_pack_start(GTK_BOX(mvbox), message_frame, TRUE, TRUE, 0);

  /* Message View */  

  mview = a_GtkMessageview_create(xetpan, message_frame, TRUE);

  uicompose->header_table = table;
  uicompose->from_combo = combo;
  uicompose->subject_entry = entry;
  uicompose->messageview = mview;

  a_GtkMessageview_show(mview);

  /* Populate FROM: combo box */

  GtkComposewindow_populate_from_combo(uicompose);

  GtkComposewindow_set_default_encoding(uicompose, PREFS_DEFAULT_CHARSET());
  /* Signals */

  g_signal_connect(G_OBJECT(toplevel),
		   "delete-event",
		   G_CALLBACK(GtkComposewindow_delete_cb),
		   uicompose);

  g_signal_connect(G_OBJECT(uicompose->messageview->text),
		   "populate-popup",
		   G_CALLBACK(GtkComposewindow_text_view_popup_cb),
		   uicompose);

  gtk_widget_show(toplevel);

  return uicompose;
}

void
a_GtkComposewindow_free(XePGTKComposeWindow *uicompose)
{
  GList *iter;

  A_LOGGING_TRACE;
  
  if (uicompose->type != XETPAN_UI_GTK)
    {
      A_LOGGING_SYSTEM_ERROR("Attempt to close a non Gtk+ UI with "
			     "a Gtk+ function");
      return;
    }
  a_UI_messageview_close((XePUIMessageView *) uicompose->messageview);

  a_GtkToolbar_composewindow_free(uicompose);

  for (iter = uicompose->headers; iter; iter = iter->next)
    {
      x_free(iter->data); /* Content are only pointers to widgets */
    }
  g_list_free(uicompose->headers);

  x_free(uicompose->sent_folder);

  gtk_widget_destroy((GtkWidget *) uicompose->toplevel);
  x_free(uicompose);
}

static GtkWidget *
GtkComposewindow_menu_create(XePGTKComposeWindow *uicompose)
{
  int i;
  GtkAccelGroup *accel_group;
  GtkItemFactoryEntry entry;
  GtkItemFactoryEntry entries[] = {
    { N_("/_File"), NULL, NULL, 0, "<Branch>" },
    { N_("/File/Send"), NULL, GtkComposewindow_file_send_cb, 0,
      "<StockItem>", GTK_STOCK_JUMP_TO},
    { N_("/File/Send later..."), NULL, GtkComposewindow_file_send_cb, 1, NULL},
    { N_("/File/Save as draft"), NULL, GtkComposewindow_file_draft_cb, 0, NULL},
    { N_("/File/_Close"), "<Control>W", GtkComposewindow_file_close_cb, 0,
      "<StockItem>", GTK_STOCK_CLOSE},
    { N_("/File/_Quit"), "<Control>Q", GtkComposewindow_file_close_cb, 1,
      "<StockItem>", GTK_STOCK_QUIT},
    { N_("/_Edit"), NULL, NULL, 0, "<Branch>" },
    { N_("/Edit/Add _To:"), NULL, GtkComposewindow_edit_add_header_cb, 0,
      NULL}, 
    { N_("/Edit/Add _Cc:"), NULL, GtkComposewindow_edit_add_header_cb, 1,
      NULL}, 
    { N_("/Edit/Add _Bcc:"), NULL, GtkComposewindow_edit_add_header_cb, 2,
      NULL}, 
    { N_("/Edit/Add _other header"), NULL, GtkComposewindow_edit_add_header_cb, 3,
      NULL}, 
    { "/Edit/sep1", NULL, NULL, 0, "<Separator>"},
    { N_("/Edit/Character _Encoding"), NULL, NULL, 0, "<Branch>" },
    { N_("/_Windows"), NULL, NULL, 0, "<Branch>" },
    { N_("/_Tools"), NULL, NULL, 0, "<Branch>" },
    { N_("/Tools/_Log Viewer"), NULL, NULL, 0, NULL},
    { N_("/_Help"), NULL, NULL, 0, "<Branch>" },
    { N_("/Help/_Contents"), "F1", NULL, 0, "<StockItem>", GTK_STOCK_HELP},
    { N_("/Help/_About"), NULL, NULL, 0, NULL}};

  accel_group = gtk_accel_group_new();
  uicompose->menu_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<Compose>", 
				      accel_group);
  gtk_item_factory_set_translate_func(uicompose->menu_factory, gettext, 
				      NULL, NULL);
  gtk_item_factory_create_items(uicompose->menu_factory, 
				sizeof(entries)/sizeof(entries[0]),
				entries, uicompose);

  entry.accelerator = NULL;
  entry.extra_data = NULL;
  for (i = 0; i < sizeof(charset_menu)/sizeof(charset_menu[0]); i++)
    {
      entry.path = charset_menu[i].path;
      entry.item_type = charset_menu[i].type;
      if (charset_menu[i].type == NULL || i == 0)
	{
	  entry.callback = GtkComposewindow_edit_charset_cb;
	  entry.callback_action = charset_menu[i].charset_index;
	  /* First charset_menu path is the radioitems leader */
	  if (i)
	    entry.item_type = charset_menu[0].path;
	}
      else
	{
	  entry.callback = NULL;
	  entry.callback_action = 0;
	}
      gtk_item_factory_create_item(uicompose->menu_factory, &entry, uicompose, 1);
    }
  gtk_window_add_accel_group(GTK_WINDOW(uicompose->toplevel), accel_group);
  return gtk_item_factory_get_widget(uicompose->menu_factory, "<Compose>");
}

static void 
GtkComposewindow_set_default_encoding(XePGTKComposeWindow *uicompose,
				      const char *charset)
{
  GtkWidget *widget;
  int i;

  if (charset == NULL)
    return;

  i = 1;
  for (i = 1;
       (xetpan_charsets[i].charset &&
	strcasecmp(xetpan_charsets[i].charset, charset));
       i++)
    {
      /* emtpy on purpose */
    }

  if (xetpan_charsets[i].charset == NULL)
    return;

  widget = gtk_item_factory_get_item_by_action(uicompose->menu_factory, i);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), TRUE);
}

static void
GtkComposewindow_populate_from_combo(XePGTKComposeWindow *uicompose)
{
  GList *list = NULL;
  GList *list_item = NULL;
  clistiter *iter;
  clist *identities = uicompose->xetpan->config->configuration->identities;
  XePIdentity *identity;
  gchar *from;

  for (iter = clist_begin(identities); iter; iter = iter->next)
    {
      identity = (XePIdentity *) iter->data;
      from = g_strdup_printf("%s <%s>",
			     (identity->name ? identity->name : ""),
			     (identity->email ? identity->email : ""));
      list = g_list_append(list, from);
    }

  if (list)
    {
      gtk_combo_set_popdown_strings(GTK_COMBO(uicompose->from_combo), list);
      for (list_item = list; list_item; list_item = list_item->next)
	{
	  g_free(list_item->data);
	}
      g_list_free(list);
    }
}

static void
GtkComposewindow_text_view_popup_cb(GtkTextView *text_view,
				    GtkMenu *menu,
				    gpointer data)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  GtkWidget *menu_item;
  
  menu_item = gtk_menu_item_new_with_label(_("Show message source"));
  g_signal_connect(G_OBJECT(menu_item), "activate",
		   G_CALLBACK(GtkComposewindow_show_source_cb),
		   uicompose);
  gtk_menu_append(menu, menu_item);
  gtk_widget_show(menu_item);

  menu_item = gtk_menu_item_new_with_label(_("Save message source as..."));
  g_signal_connect(G_OBJECT(menu_item), "activate",
		   G_CALLBACK(GtkComposewindow_save_source_as_cb),
		   uicompose);
  gtk_menu_append(menu, menu_item);
  gtk_widget_show(menu_item);

}

static gboolean
GtkComposewindow_delete_cb(GtkWidget *widget,
			   GdkEvent *event,
			   gpointer data)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  XetPan *xetpan = (XetPan *) uicompose->xetpan;

  A_LOGGING_TRACE;

  if (a_UI_composewindow_close((XePUIComposeWindow *) uicompose))
    {
      a_XetPan_quit(xetpan);
    }

  return FALSE;
}

static void
GtkComposewindow_file_send_cb(gpointer data,
			      guint action,
			      GtkWidget *widget)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;

  switch (action)
    {
    case 0:
      a_GtkComposewindow_send(uicompose);
      break;
    case 1:
      /* TODO: Ask for when to send */
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Unrecognised send action type");
    }
}

static void
GtkComposewindow_file_draft_cb(gpointer data,
			       guint action,
			       GtkWidget *widget)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  
  A_LOGGING_TRACE;
}

static void
GtkComposewindow_file_close_cb(gpointer data,
			       guint action,
			       GtkWidget *widget)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  XetPan *xetpan = (XetPan *) uicompose->xetpan;

  A_LOGGING_TRACE;

  if (action || a_UI_composewindow_close((XePUIComposeWindow *)uicompose))
    {
      a_XetPan_quit(xetpan);
    }
}

static void
GtkComposewindow_edit_add_header_cb(gpointer data,
				    guint action,
				    GtkWidget *widget)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  switch (action)
    {
    case 0:
      a_GtkComposewindow_add_to_header(uicompose, NULL);
      break;
    case 1:
      a_GtkComposewindow_add_cc_header(uicompose, NULL);
      break;
    case 2:
      a_GtkComposewindow_add_bcc_header(uicompose, NULL);
      break;
    default:
      a_GtkComposewindow_add_other_header(uicompose, NULL, NULL);
    }
}

static void
GtkComposewindow_edit_charset_cb(gpointer data,
				 guint action,
				 GtkWidget *widget)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  const char *charset = (xetpan_charsets[action]).charset;

  x_free(uicompose->messageview->default_charset);
  uicompose->messageview->default_charset = strdup(charset);  
}

static void
GtkComposewindow_show_source_cb(GtkMenuItem *menu_item, gpointer data)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;
  char *filename;
  char *charset = NULL;
  char *display;

  GtkComposewindow_build_message(uicompose);
  filename = a_Message_write_to_temp_file(uicompose->messageview->message);

  if (filename)
    {
      display = a_GtkUtils_make_display_name(uicompose->toplevel);
      a_GtkUtils_file_viewer_new(display, _("Message source"),
				 NULL, 0, 0, filename, NULL, 0, charset);
      g_free(display);
      
      unlink(filename);
      x_free(filename);
    }
}

static void
GtkComposewindow_save_source_as_cb(GtkMenuItem *menu_item, gpointer data)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;

  GtkComposewindow_build_message(uicompose);

  /* TODO: create a file name from the subject, the sender and the date:
   * %s_%f_%d.txt e.g.:[XetPan]_Plan-melvin_hadasht-030602.txt
   * Use the message template functions for that.
   */
  a_GtkUtils_file_selection(_("Save message source as..."),
			    uicompose->toplevel,
			    "email.txt",
			    GtkComposewindow_save_source_as,
			    uicompose->messageview->message);
}

static int
GtkComposewindow_save_source_as(void *amessage, const char *filename)
{
  XePMessage *message = (XePMessage *) amessage;
  FILE *fstr;
  int r = -1;

  fstr = fopen(filename, "w");
  if (fstr)
    {
      a_Message_write_to_file(message, fstr);
      fclose(fstr);
      r = 0;
    }

  return r;
}

void
a_GtkComposewindow_add_header(XePGTKComposeWindow *uicompose, 
			      XePMessageHeaderType header_type,
			      const char *header_name,
			      const char *header_value)
{
  GtkComposewindow_add_header(uicompose, -1, header_type, header_name,
			      header_value);
}

void
a_GtkComposewindow_add_to_header(XePGTKComposeWindow *uicompose,
				 const char *value)
{
  GtkComposewindow_add_header(uicompose, -1, XEP_MAIL_HEADER_TO, NULL,
			      value);
}

void
a_GtkComposewindow_add_cc_header(XePGTKComposeWindow *uicompose,
				 const char *value)
{
  GtkComposewindow_add_header(uicompose, -1, XEP_MAIL_HEADER_CC, NULL,
			      value);
}

void
a_GtkComposewindow_add_bcc_header(XePGTKComposeWindow *uicompose,
				  const char *value)
{
  GtkComposewindow_add_header(uicompose, -1, XEP_MAIL_HEADER_BCC, NULL,
			      value);
}

void
a_GtkComposewindow_add_other_header(XePGTKComposeWindow *uicompose,
				    const char *other_header,
				    const char *value)
{
  GtkComposewindow_add_header(uicompose, -1, XEP_MAIL_HEADER_OTHER, 
			      other_header,
			      value);
}

/*! Insert a header at line row. row = 0 is the first line
 *  after the From: header. So row corresponds to table row
 *  attachments row + 1 to row + 2 and to row'th item in 
 *  the XePGTKComposeWindow::headers list
 *  Append happens when the to be inserted row does not exist, 
 *  i.e. when row + 1 >= table::nrows (or when row == -1)
 */
static void
GtkComposewindow_add_header(XePGTKComposeWindow *uicompose,
			    int row,
			    XePMessageHeaderType header_type,
			    const char *other_header,
			    const char *header_value)
{
  const char *header_name = NULL;

  if (header_type == XEP_MAIL_HEADER_OTHER)
    {
      if (other_header)
	{
	  header_name = other_header;
	}
      else
	{
	  /* TODO: Interactive query of the header to be added */
	  
	}
    }
  else
    {
      header_name = a_Message_header_get_name_from_type(header_type);
    }

  GtkComposewindow_add_header_real(uicompose, row, header_name, header_value);
}

static void
GtkComposewindow_add_header_real(XePGTKComposeWindow *uicompose,
				 int row,
				 const char *header_name,
				 const char *header_value)
{
  GtkTable *table = GTK_TABLE(uicompose->header_table);
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *add;
  GtkWidget *remove;
  XePGTKComposeWindowHeader *header;
  XePGTKComposeWindowHeader *prev_header;
  gchar *prev_name;
  gchar *prev_val;
  const gchar *name;
  const gchar *val;
  GList *iter;
  GList *new_iter;
  int list_index;
  int table_index;

  if (row < 0)
    {
      list_index = g_list_length(uicompose->headers);
    }
  else
    {
      list_index = row;
    }

  /* Add an empty header row at the end */
  table_index = table->nrows;

  label = a_GtkUtils_label_new_bold(NULL);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_widget_show(label);

  entry = gtk_entry_new();
  gtk_widget_show(entry);

  add = a_GtkUtils_button_new_from_stock(GTK_STOCK_ADD, NULL, FALSE,
					 GTK_ICON_SIZE_SMALL_TOOLBAR);
  gtk_button_set_relief(GTK_BUTTON(add), GTK_RELIEF_NONE);
  gtk_widget_show(add);

  remove = a_GtkUtils_button_new_from_stock(GTK_STOCK_REMOVE, NULL, FALSE,
					    GTK_ICON_SIZE_SMALL_TOOLBAR);
  gtk_button_set_relief(GTK_BUTTON(remove), GTK_RELIEF_NONE);
  gtk_widget_show(remove);

  /* Create the corresponding header structure */

  header = g_new(XePGTKComposeWindowHeader, 1);
  header->label     = label;
  header->entry     = entry;
  header->add       = add;
  header->remove    = remove;
  header->uicompose = uicompose;
  header->index     = g_list_length(uicompose->headers);;

  uicompose->headers = g_list_append(uicompose->headers, header);

  /* Signals to add a similar header or to remove this one */

  g_signal_connect(G_OBJECT(add), "clicked",
		   G_CALLBACK(GtkComposewindow_add_header_cb),
		   header);
  
  g_signal_connect(G_OBJECT(remove), "clicked",
		   G_CALLBACK(GtkComposewindow_remove_header_cb),
		   header);

  /* Add header row at the end of the table */
  
  gtk_table_attach(table, header->label,  0, 1, table_index, table_index + 1,
		   GTK_SHRINK | GTK_FILL,
		   GTK_SHRINK,
		   12, 0);
  gtk_table_attach(table, header->entry,  1, 2, table_index, table_index + 1,
		   GTK_FILL | GTK_EXPAND,
		   GTK_SHRINK,
		   0, 0);
  gtk_table_attach(table, header->add,    2, 3, table_index, table_index + 1,
		   GTK_SHRINK,
		   GTK_SHRINK,
		   0, 0);
  gtk_table_attach(table, header->remove, 3, 4, table_index, table_index + 1,
		   GTK_SHRINK,
		   GTK_SHRINK,
		   0, 0);
  
  /* Get the iter corresponding to the inserted row */
  
  new_iter = g_list_nth(uicompose->headers, list_index);
  
  /* Loop from the end until the inserted row and move
   * rows up one line.
   * 'header' is already the last row
   */

  prev_header = header;

  for (iter = g_list_last(uicompose->headers);
       iter && iter != new_iter;
       iter = iter->prev)
    {
      header = prev_header;
      prev_header = (XePGTKComposeWindowHeader *) iter->prev->data;
      
      /* Move labels and entries values one row up */

      prev_name = g_strdup(gtk_label_get_text(GTK_LABEL(prev_header->label)));
      prev_val  = g_strdup(gtk_entry_get_text(GTK_ENTRY(prev_header->entry)));
      name = gtk_label_get_text(GTK_LABEL(header->label));
      val  = gtk_entry_get_text(GTK_ENTRY(header->entry));

      gtk_label_set_text(GTK_LABEL(prev_header->label), name);
      gtk_entry_set_text(GTK_ENTRY(prev_header->entry), val);
      gtk_label_set_text(GTK_LABEL(header->label), prev_name);
      gtk_entry_set_text(GTK_ENTRY(header->entry), prev_val);

      g_free(prev_name);
      g_free(prev_val);
    }

  /* Set the label and the entry values of the new row */

  gtk_label_set_text(GTK_LABEL(prev_header->label), header_name);
  if (header_value)
    {
      gtk_entry_set_text(GTK_ENTRY(prev_header->entry), header_value);
    }  
}

static void
GtkComposewindow_remove_header(XePGTKComposeWindowHeader *header)
{
  XePGTKComposeWindow *uicompose = header->uicompose;
  GtkTable *table = GTK_TABLE(uicompose->header_table);
  int index = header->index;
  GList *iter;
  const char *header_name;

  header_name = gtk_label_get_text(GTK_LABEL(header->label));
  a_Message_header_remove_by_name(uicompose->messageview->message, header_name);

  gtk_container_remove(GTK_CONTAINER(table), header->label);
  gtk_container_remove(GTK_CONTAINER(table), header->entry);
  gtk_container_remove(GTK_CONTAINER(table), header->add);
  gtk_container_remove(GTK_CONTAINER(table), header->remove);

  uicompose->headers = g_list_remove(uicompose->headers, header);
  g_free(header);

  for (iter = g_list_last(uicompose->headers); iter; iter = iter->prev)
    {
      header = (XePGTKComposeWindowHeader *) iter->data;
      if (header->index > index)
	{
	  header->index--;

	}
      else
	{
	  break;
	}
    }

}

static void
GtkComposewindow_add_header_cb(GtkWidget *button, gpointer data)
{
  XePGTKComposeWindowHeader *header = (XePGTKComposeWindowHeader *) data;
  const char *header_name;

  header_name = gtk_label_get_text(GTK_LABEL(header->label));

  GtkComposewindow_add_header_real(header->uicompose, 
				   header->index + 1,
				   header_name,
				   NULL);
}

static void
GtkComposewindow_remove_header_cb(GtkWidget *button, gpointer data)
{
  XePGTKComposeWindowHeader *header = (XePGTKComposeWindowHeader *) data;

  GtkComposewindow_remove_header(header);
}

void
a_GtkComposewindow_set_message(XePGTKComposeWindow *uicompose, 
			       XePMessage *message)
{
  a_GtkMessageview_mime_view_show(uicompose->messageview,
				  message);
  a_GtkComposewindow_list_headers(uicompose);
}

void
a_GtkComposewindow_set_body_text(XePGTKComposeWindow *uicompose,
				 const char *text,
				 size_t text_size,
				 const char *text_charset)
{
  XePGTKMessageView *mview = (XePGTKMessageView *)uicompose->messageview;
  GtkTextIter iter;
  char *quoted_text = NULL;

  gtk_text_buffer_get_end_iter(mview->buffer, &iter);

  /* TODO: The initial text should be created by a template function
   */

  /* TODO: use text buffer marks and tags to improve the rendering
   * of the quoted text
   */
  a_Utils_quote_text_for_reply(text,
			       text_size,
			       text_charset,
			       &quoted_text);

  if (quoted_text)
    {
      a_GtkUtils_text_buffer_convert_insert(mview->buffer, &iter,
					    quoted_text, strlen(quoted_text),
					    text_charset);
      x_free(quoted_text);
    }
}

void
a_GtkComposewindow_list_headers(XePGTKComposeWindow *uicompose)
{
  char *header_value;
  mailmessage *msg;
  int i;
  int headers[] = {
    XEP_MAIL_HEADER_TO,
    XEP_MAIL_HEADER_CC,
  };
  char *templates[] = {
    "%t <%T>",
    "%c <%C>", /* This is a XetPan's CC template */
    NULL};

  /* TODO: we should not use the mailmessage::single_fields
   * but only mailmime::fields, so we can get rid
   * of mailmessage and use only mailmime
   */
  msg = uicompose->messageview->message->msg;

  header_value = a_Message_message_formatted_get_header(msg,
							XEP_MAIL_HEADER_FROM,
							NULL,
							"%f <%F>",
							GTK_DEST_CHARSET);
  if (header_value)
    {
      gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(uicompose->from_combo)->entry),
			 header_value);
      x_free(header_value);
    }


  for (i = 0;i < sizeof(headers)/sizeof(headers[0]); i++)
    {
      header_value = a_Message_message_formatted_get_header(msg, 
							    headers[i],
							    NULL,
							    templates[i],
							    GTK_DEST_CHARSET);
      if (header_value)
	{      
	  GtkComposewindow_add_header(uicompose, -1,
				      headers[i], NULL,
				      header_value);
	  free(header_value);
	}
    }

  header_value = a_Message_message_formatted_get_header(msg,
							XEP_MAIL_HEADER_SUBJECT,
							NULL,
							"%s",
							GTK_DEST_CHARSET);
  if (header_value)
    {
      gtk_entry_set_text(GTK_ENTRY(uicompose->subject_entry), header_value);
      free(header_value);
    }
}

void
a_GtkComposewindow_send(XePGTKComposeWindow *uicompose)
{
  XetPan *xetpan = uicompose->xetpan;
  clist *identities = xetpan->config->configuration->identities;
  const char *from;
  XePIdentity *identity;
  XePMTA *mta = NULL;
  const char *sent_ref = NULL;

  GtkComposewindow_build_message(uicompose);

  from = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(uicompose->from_combo)->entry));
  identity = a_Identity_lookup_by_email(identities, from);

  if (identity)
    {
      const char *mta_ref;
      mta_ref = a_Identity_get_mta_ref(identity);
      /* if (mta_ref == NULL) then choose first registered MTA */
      mta = a_MTA_lookup_by_ref(mta_ref);
      if (uicompose->sent_folder == NULL)
	{
	  sent_ref = a_Identity_get_sent_ref(identity);
	  if (sent_ref)
	    {
	      uicompose->sent_folder = strdup(sent_ref);
	      if (uicompose->sent_folder == NULL)
		{
		  A_LOGGING_ERROR("strdup(): %s", strerror(errno));
		}
	    }
	}
    }

  if (mta == NULL)
    {
      A_LOGGING_ERROR(_("No Mail Transfer Agent (MTA) was found."));
      return;
    }
  gtk_widget_hide(uicompose->toplevel);
   
  a_MTA_send_message(mta, uicompose->messageview->message,
		     GtkComposewindow_sent_cb, uicompose,
		     uicompose->sent_folder);
}

static void
GtkComposewindow_sent_cb(void *auicompose, void *aresult)
{
  int result = (int) aresult;
  XePUIComposeWindow *uicompose = (XePUIComposeWindow *) auicompose;
  XetPan *xetpan = uicompose->any.xetpan;

  if (uicompose->any.sent_folder)
    a_UI_update_folder(uicompose->any.sent_folder);

  if (result < 0)
    {
      gtk_widget_show(((XePGTKComposeWindow *) uicompose)->toplevel);
    }
  else if (a_UI_composewindow_close(uicompose))
    {
      a_XetPan_quit(xetpan);
    }
}

static void
GtkComposewindow_build_message(XePGTKComposeWindow *uicompose)
{
  char *text;
  const char *header_name;
  const char *header_value;
  const gchar *subject;
  const gchar *from;
  XePMessage *message = uicompose->messageview->message;
  XePGTKComposeWindowHeader *header;
  GList *iter;

  /* Remove all headers that can be added from the UI */

  for (iter = uicompose->headers; iter; iter = iter->next)
    {
      header = (XePGTKComposeWindowHeader *) iter->data;
      header_name = gtk_label_get_text(GTK_LABEL(header->label));
      header_value = gtk_entry_get_text(GTK_ENTRY(header->entry));
      a_Message_header_remove_by_name(message, header_name);
    }

  for (iter = uicompose->headers; iter; iter = iter->next)
    {
      header = (XePGTKComposeWindowHeader *) iter->data;
      header_name = gtk_label_get_text(GTK_LABEL(header->label));
      header_value = gtk_entry_get_text(GTK_ENTRY(header->entry));
      a_Message_header_add_by_name(message, header_name, header_value, 
				   GTK_DEST_CHARSET);
    }

  from = 
    gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(uicompose->from_combo)->entry));
  a_Message_header_set_from(message, from, GTK_DEST_CHARSET);

  subject = gtk_entry_get_text(GTK_ENTRY(uicompose->subject_entry));
  a_Message_header_set_subject(message, subject, GTK_DEST_CHARSET);

  if (message->selected_mime == message->default_mime)
    {
      text = a_GtkMessageview_get_text(uicompose->messageview);
      /* FIXME: Check for return value */
      a_GtkMessageview_default_part_set_text(uicompose->messageview,
					     text, (text ? strlen(text) : 0));
      g_free(text);
    }
}
#endif
