/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtk.h>

#include <clist.h>

#include "intl.h"

#include "x_alloc.h"
#include "defines.h"
#include "gtk-mb-editor.h"
#include "logging.h"
#include "identity.h"
#include "mailbox.h"
#include "processing.h"
#include "gtk-utils.h"
#include "ui.h"
#include "mta.h"
#include "utils.h"
#include "gtk-mta-editor.h"

#define ID_TRIPLET_COPY(triplet)    ((XePIdentity *) (triplet)->data1)
#define ID_TRIPLET_ORIG(triplet)    ((XePIdentity *) (triplet)->data2)
#define ID_TRIPLET_FLAGS(triplet) ((unsigned int) (triplet)->data3)

#define ID_CHANGES_IDEDITOR(changes) ((XePGtkIdEditor *) changes->name)
#define ID_CHANGES_TRIPLET(changes) ((XePUtilsTriplet *) changes->value)

typedef enum
  {
    XEP_IDEDITOR_IDENTITY,
    XEP_IDEDITOR_SIGNATURES,
    XEP_IDEDITOR_MAILBOXES,
    XEP_IDEDITOR_MTAS,
    XEP_IDEDITOR_DEFAULTS
  } XePGtkIdEditorPage;

enum
  {
    XEP_IDEDITOR_NAME_CHANGED       = 1 << 1,
    XEP_IDEDITOR_EMAIL_CHANGED      = 1 << 2,
    XEP_IDEDITOR_ORG_CHANGED        = 1 << 3,
    XEP_IDEDITOR_REPLY_TO_CHANGED   = 1 << 4,
    XEP_IDEDITOR_SIGNATURES_CHANGED = 1 << 5,
    XEP_IDEDITOR_MAILBOXES_CHANGED  = 1 << 6,
    XEP_IDEDITOR_MTAS_CHANGED       = 1 << 7,
    XEP_IDEDITOR_DRAFTS_CHANGED     = 1 << 8,
    XEP_IDEDITOR_QUEUE_CHANGED      = 1 << 9,
    XEP_IDEDITOR_SENT_CHANGED       = 1 << 10,
    XEP_IDEDITOR_TEMPLATES_CHANGED  = 1 << 11,
    XEP_IDEDITOR_EXPANDED           = 1 << 12,
    XEP_IDEDITOR_NEW                = 1 << 13,
    XEP_IDEDITOR_SELECT             = 1 << 14
  };

#define XEP_IDEDITOR_IDENTITY_CHANGED (XEP_IDEDITOR_NAME_CHANGED  | \
			         XEP_IDEDITOR_EMAIL_CHANGED | \
                                 XEP_IDEDITOR_ORG_CHANGED   | \
			         XEP_IDEDITOR_REPLY_TO_CHANGED)

#define XEP_IDEDITOR_DEFAULTS_CHANGED (XEP_IDEDITOR_DRAFTS_CHANGED   | \
                                       XEP_IDEDITOR_QUEUE_CHANGED    | \
                                       XEP_IDEDITOR_SENT_CHANGED     | \
                                       XEP_IDEDITOR_TEMPLATES_CHANGED)

#define XEP_IDEDITOR_ALL_CHANGED (XEP_IDEDITOR_IDENTITY_CHANGED | \
                                  XEP_IDEDITOR_SIGNATURES_CHANGED | \
                                  XEP_IDEDITOR_MAILBOXES_CHANGED  | \
                                  XEP_IDEDITOR_MTAS_CHANGED       | \
                                  XEP_IDEDITOR_DEFAULTS_CHANGED)

typedef struct XePGtkIdEditor_s XePGtkIdEditor;
struct XePGtkIdEditor_s
{
  GtkWidget *widget;
  GtkWidget *buttons;

  clist *identity_list;

  /* Identity bar */
  GtkWidget *id_option_menu;
  GtkWidget *id_new;
  GtkWidget *id_delete;

  /* Pages */
  GtkWidget *notebook;

  /* ID page */
  GtkWidget *id_name;
  GtkWidget *id_email;
  GtkWidget *id_org;
  GtkWidget *id_reply_to;
  GtkWidget *id_default;

  /* Signatures page */
  XePGtkUtilsTree *sigtree;

  /* Mailboxes page */
  XePGtkUtilsTree *mbtree;
  XePGtkMbEditor *mbeditor;

  /* MTAs page */
  GtkWidget *mtas_optmenu;
  GtkWidget *mtas_props;
  GtkWidget *mtas_remove;
  XePGtkMTAEditor *mtaeditor;

  /* Defaults page*/
  GtkWidget *def_drafts;
  GtkWidget *def_queue;
  GtkWidget *def_sent;
  GtkWidget *def_templates;

  XePUtilsTriplet *current_triplet;
  GtkWidget *current_id_label;

  guint id_count;
  guint opened_editors;
};

/* Functions declarations */

static int
GtkIdEditor_create(XePUIMain *uimain);

static void
GtkIdEditor_hide(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_destroy_cb(GObject *adialog, gpointer data);

static void
GtkIdEditor_cancel_cb(GtkButton *button, gpointer data);

static void
GtkIdEditor_ok_cb(GtkButton *button, gpointer data);

static void
GtkIdEditor_identity_finalize(XePUtilsPair *changes);

static void
GtkIdEditor_identity_free_finalize(XePIdentity *identity, void *achanges);

static void
GtkIdEditor_ok_finalize(XePGtkIdEditor *ideditor);

static gboolean
GtkIdEditor_delete_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data);

static void
GtkIdEditor_transfer_changes(XePUtilsPair *changes);

static void
GtkIdEditor_transfer_changes_finalize_clone(clist *mailboxes, void *achanges);

static void
GtkIdEditor_apply_changes_to_copy(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_update_pages(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_set_opened_editor(XePGtkIdEditor *ideditor, 
			      XePGtkIdEditorPage page,
			      gboolean open);

static XePGtkUtilsTree *
GtkIdEditor_append_tree_page(XePGtkIdEditor *ideditor,
			     const char *title,
			     const char *help,
			     clist *list);
static void
GtkIdEditor_add_coordinates_page(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_add_signatures_page(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_add_mailboxes_page(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_add_mtas_page(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_add_defaults_page(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_identity_add_cb(GtkButton *button, void *aideditor);

static void
GtkIdEditor_identity_remove_cb(GtkButton *button, void *aideditor);

static void
GtkIdEditor_identity_set_default_cb(GtkButton *button, void *aideditor);

static void
GtkIdEditor_identity_set_changed(XePGtkIdEditor *ideditor, 
				 unsigned int set, unsigned int unset);

static int
GtkIdEditor_identities_new_copy(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_identities_free_copy(XePGtkIdEditor *ideditor);
  
static void *
GtkIdEditor_mailboxes_remove_cb(XePGtkUtilsTree *mbtree, void *aideditor);

#if 0
static void *
GtkIdEditor_mailboxes_to_top_cb(XePGtkUtilsTree *mbtree, void *aideditor);
#endif

static void *
GtkIdEditor_mailboxes_refresh_cb(XePGtkUtilsTree *idtree, void *aideditor);

static void
GtkIdEditor_init(XePGtkIdEditor *ideditor, XePUtilsTriplet *show_triplet);

static void GtkIdEditor_identity_activate_cb(GtkMenuItem *menuitem, 
					     gpointer aideditor);

static void
GtkIdEditor_entry_changed(GtkEditable *editable, gpointer aideditor);

static void *
GtkIdEditor_sig_add_cb(XePGtkUtilsTree *sigtree, void *aideditor);

static void
GtkIdEditor_sig_add_real_cb(void *aideditor, int response, gchar *signature);

static void *
GtkIdEditor_sig_remove_cb(XePGtkUtilsTree *sigtree, void *aideditor);

static void *
GtkIdEditor_sig_changed_cb(XePGtkUtilsTree *sigtree, void *aideditor);

static void *
GtkIdEditor_sig_refresh_cb(XePGtkUtilsTree *sigtree, void *aideditor);

static void
GtkIdEditor_mailboxes_cancel_cb(XePGtkMbEditor *mbeditor, void *aideditor);

static void *
GtkIdEditor_mailboxes_add_cb(XePGtkUtilsTree *mbtree, void *aideditor);

static void
GtkIdEditor_mailboxes_add_done_cb(XePGtkMbEditor *mbeditor, void *aideditor);

static void *
GtkIdEditor_mailboxes_edit_cb(XePGtkUtilsTree *mbtree, void *aideditor);

static void
GtkIdEditor_mailboxes_edit_done_cb(XePGtkMbEditor *mbeditor, void *aideditor);

static void *
GtkIdEditor_mailboxes_changed_cb(XePGtkUtilsTree *mbtree, void *aideditor);

static void 
GtkIdEditor_mtas_update_menu(XePGtkIdEditor *ideditor, const char *def_mta);

static const char *
GtkIdEditor_mtas_get_selected(XePGtkIdEditor *ideditor);

static void
GtkIdEditor_mta_changed_cb(GtkOptionMenu *optmenu, gpointer aideditor);

static void
GtkIdEditor_mta_edit_cb(GtkButton *button, gpointer aideditor);

static void
GtkIdEditor_mta_edit_done_cb(XePGtkMTAEditor *mtaeditor, gpointer aideditor);

static void
GtkIdEditor_mta_add_cb(GtkButton *button, gpointer aideditor);

static void
GtkIdEditor_mta_add_done_cb(XePGtkMTAEditor *mtaeditor, gpointer aideditor);

static void
GtkIdEditor_mta_remove_cb(GtkButton *button, gpointer aideditor);

static void
GtkIdEditor_mta_cancel_cb(XePGtkMTAEditor *mtaeditor, gpointer aideditor);


/* Functions definitions */

static int
GtkIdEditor_create(XePUIMain *uimain)
{
  GtkWidget *vbox;
  GtkWidget *hbuttonbox;
  GtkWidget *cancelbutton;
  GtkWidget *okbutton;
  XePGtkIdEditor *ideditor;
  GtkWidget *label;
  GtkWidget *newbutton;
  GtkWidget *deletebutton;
  GtkWidget *hbox;
  GtkWidget *option;

  A_LOGGING_TRACE;

  ideditor = x_alloc(1, XePGtkIdEditor);
  if (ideditor == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      return -1;
    }
  
  ideditor->current_triplet = NULL;
  ideditor->current_id_label = NULL;
  ideditor->mbeditor = NULL;
  ideditor->opened_editors = 0;

  ideditor->widget = a_GtkUtils_transient_new(uimain->toplevel,
					      _("XetPan Identities Settings"),
					      -1, -1);

  vbox = gtk_vbox_new(FALSE, 12);
  gtk_container_add(GTK_CONTAINER(ideditor->widget), vbox);
  gtk_widget_show(vbox);

  /* Identities hbox */

  hbox = gtk_hbox_new(FALSE, 12);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show(hbox);

  label = a_GtkUtils_label_new_bold_large(_("Identity:"));
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);

  option = gtk_option_menu_new();
  gtk_box_pack_start(GTK_BOX(hbox), option, TRUE, TRUE, 0);
  gtk_widget_show(option);

  ideditor->id_option_menu = option;

  hbuttonbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(hbuttonbox), GTK_BUTTONBOX_EDGE);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbuttonbox), 12);
  gtk_widget_show(hbuttonbox);
  gtk_box_pack_start(GTK_BOX(hbox), hbuttonbox, FALSE, FALSE, 0);

  newbutton = gtk_button_new_with_mnemonic(_("New"));
  gtk_container_add(GTK_CONTAINER(hbuttonbox), newbutton);
  gtk_widget_show(newbutton);

  ideditor->id_new = newbutton;

  deletebutton = gtk_button_new_with_mnemonic(_("Delete"));
  gtk_container_add(GTK_CONTAINER(hbuttonbox), deletebutton);
  gtk_widget_show(deletebutton);

  ideditor->id_delete = deletebutton;

  /* Settings notebook */

  ideditor->notebook = gtk_notebook_new();
  gtk_widget_show(ideditor->notebook);
  gtk_box_pack_start(GTK_BOX(vbox), ideditor->notebook, TRUE, TRUE, 0);

  GtkIdEditor_add_coordinates_page(ideditor);
  GtkIdEditor_add_signatures_page(ideditor);
  GtkIdEditor_add_mailboxes_page(ideditor);
  GtkIdEditor_add_mtas_page(ideditor);
  GtkIdEditor_add_defaults_page(ideditor);

  ideditor->buttons = hbuttonbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(hbuttonbox), GTK_BUTTONBOX_END);
  gtk_box_set_spacing(GTK_BOX(hbuttonbox), 12);
  gtk_box_pack_start(GTK_BOX(vbox), hbuttonbox, FALSE, FALSE, 0);
  gtk_widget_show(hbuttonbox);

  cancelbutton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_container_add(GTK_CONTAINER(hbuttonbox), cancelbutton);
  gtk_widget_show(cancelbutton);

  okbutton = gtk_button_new_from_stock(GTK_STOCK_OK);
  gtk_container_add(GTK_CONTAINER(hbuttonbox), okbutton);
  gtk_widget_show(okbutton);

  uimain->ideditor = ideditor;

  /* Identities Editor Signals */

  g_signal_connect(G_OBJECT(newbutton), "clicked",
		   G_CALLBACK(GtkIdEditor_identity_add_cb),
		   ideditor);
  g_signal_connect(G_OBJECT(deletebutton), "clicked",
		   G_CALLBACK(GtkIdEditor_identity_remove_cb),
		   ideditor);
  g_signal_connect(G_OBJECT(cancelbutton), "clicked",
		   G_CALLBACK(GtkIdEditor_cancel_cb),
		   ideditor);

  g_signal_connect(G_OBJECT(okbutton), "clicked",
		   G_CALLBACK(GtkIdEditor_ok_cb),
		   ideditor);

  g_signal_connect(G_OBJECT(ideditor->widget), "delete-event",
                   G_CALLBACK(GtkIdEditor_delete_event_cb),
                   ideditor);

  g_signal_connect(G_OBJECT(ideditor->widget), "destroy",
		   G_CALLBACK(GtkIdEditor_destroy_cb),
		   ideditor);
  return 0;
}

static XePGtkUtilsTree *
GtkIdEditor_append_tree_page(XePGtkIdEditor *ideditor,
			     const char *title,
			     const char *help,
			     clist *list)
{
  XePGtkUtilsTree *utils_tree;
  GtkWidget *table;
  GtkWidget *label;
  
  table = gtk_table_new(4, 2, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 12);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  
  utils_tree = a_GtkUtils_tree_new(GTK_ORIENTATION_HORIZONTAL, TRUE,
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  
  if (list)
    {
      a_GtkUtils_tree_set_list(utils_tree, list);
      a_GtkUtils_tree_refresh(utils_tree);
    }
  
  gtk_table_attach(GTK_TABLE(table), XEP_GTKUTILS_TREE_WIDGET(utils_tree),
		   0, 1, 1, 2,
		   GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);

  gtk_table_attach(GTK_TABLE(table), XEP_GTKUTILS_TREE_BUTTONS(utils_tree),
		   0, 1, 2, 3,
		   GTK_FILL | GTK_EXPAND, 0, 0, 0);

  gtk_widget_show(XEP_GTKUTILS_TREE_WIDGET(utils_tree));
  gtk_widget_show(XEP_GTKUTILS_TREE_BUTTONS(utils_tree));

  label = a_GtkUtils_label_new_italic(help);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.50);
  gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4,
		   GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(label);
  
  label = gtk_label_new_with_mnemonic(title);
  gtk_widget_show(label);

  gtk_notebook_append_page(GTK_NOTEBOOK(ideditor->notebook), table, label);
  gtk_widget_show(table); 
  return utils_tree;
}

static void
GtkIdEditor_add_coordinates_page(XePGtkIdEditor *ideditor)
{
  GtkWidget *label;
  GtkWidget *table;
  int i = 0;

  table = gtk_table_new(5, 2, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 12);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);

  ideditor->id_name = 
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Name:"), NULL,
			       _("E.g.: William Riker"), 0, i);
  g_object_set_data(G_OBJECT(ideditor->id_name), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_NAME_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->id_name), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);
  i ++;

  ideditor->id_email = 
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Email:"), NULL,
			       _("E.g.: william.riker@organization.org"),
			       0, i);
  g_object_set_data(G_OBJECT(ideditor->id_email), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_EMAIL_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->id_email), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);
  i ++;

  ideditor->id_org = 
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Organization:"), NULL,
			       _("E.g.: The Riker & Riker Organization"), 
			       0, i);
  g_object_set_data(G_OBJECT(ideditor->id_org), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_ORG_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->id_org), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);
  i ++;

  ideditor->id_reply_to =
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Reply-To:"), NULL,
			       _("Explicitly request replies to be addressed "
				 "to this email address.  To be used only "
				 "if the replies are not meant to be sent to "
				 "your own email address."), 0, i);
  g_object_set_data(G_OBJECT(ideditor->id_reply_to), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_REPLY_TO_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->id_reply_to), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);

  i++;

  ideditor->id_default = 
    gtk_button_new_with_mnemonic(_("Set as Default Identity"));
  gtk_table_attach(GTK_TABLE(table), ideditor->id_default,
		   0, 2, i, i + 1, GTK_EXPAND, GTK_EXPAND, 0, 0);
  gtk_widget_show(ideditor->id_default);
		   
  /* Signal callbacks when entry is changed */

  g_signal_connect(G_OBJECT(ideditor->id_default), "clicked", 
		   G_CALLBACK(GtkIdEditor_identity_set_default_cb),
		   ideditor);

  label = gtk_label_new_with_mnemonic(_("_Identity"));
  gtk_widget_show(label);
  
  gtk_notebook_append_page(GTK_NOTEBOOK(ideditor->notebook), table, label);
  gtk_widget_show(table);

}

static void
GtkIdEditor_add_signatures_page(XePGtkIdEditor *ideditor)
{
  XePGtkUtilsTree *sigtree;
  sigtree = GtkIdEditor_append_tree_page(ideditor, _("_Signatures"),
					 _("Above is the list of "
					   "files containing signatures you "
					   "can use with this identity."),
					 NULL);

  a_GtkUtils_tree_set_add_callback(sigtree, GtkIdEditor_sig_add_cb, 
				   ideditor);
  a_GtkUtils_tree_set_remove_callback(sigtree, GtkIdEditor_sig_remove_cb, 
				      ideditor);

#if 0
  a_GtkUtils_tree_set_edit_callback(sigtree,   NULL, NULL);
  a_GtkUtils_tree_set_top_callback(sigtree,    NULL, NULL);
#endif
  a_GtkUtils_tree_set_changed_callback(sigtree, GtkIdEditor_sig_changed_cb,
				       ideditor);
  a_GtkUtils_tree_set_refresh_callback(sigtree, GtkIdEditor_sig_refresh_cb,
				       ideditor);
  ideditor->sigtree = sigtree;
}

static void
GtkIdEditor_add_mailboxes_page(XePGtkIdEditor *ideditor)
{
  XePGtkUtilsTree *mbtree;
  
  mbtree = GtkIdEditor_append_tree_page(ideditor, _("_Mailboxes"),
					_("Above is the list of mailboxes "
					  "you can use with this identity."),
					NULL);

  a_GtkUtils_tree_set_add_callback(mbtree,    
				   GtkIdEditor_mailboxes_add_cb,
				   ideditor);
  a_GtkUtils_tree_set_remove_callback(mbtree, 
				      GtkIdEditor_mailboxes_remove_cb, 
				      ideditor);
  a_GtkUtils_tree_set_edit_callback(mbtree,
				    GtkIdEditor_mailboxes_edit_cb,
				    ideditor);
#if 0
  a_GtkUtils_tree_set_top_callback(mbtree,    NULL, NULL);
#endif
  a_GtkUtils_tree_set_changed_callback(mbtree, 
				       GtkIdEditor_mailboxes_changed_cb,
				       ideditor);
  a_GtkUtils_tree_set_refresh_callback(mbtree, 
				       GtkIdEditor_mailboxes_refresh_cb,
				       ideditor);

  ideditor->mbtree = mbtree;

  ideditor->mbeditor = a_GtkMbEditor_new(ideditor->widget);
  a_GtkMbEditor_set_cancel_callback(ideditor->mbeditor,
				    GtkIdEditor_mailboxes_cancel_cb,
				    ideditor);

}

static void
GtkIdEditor_add_mtas_page(XePGtkIdEditor *ideditor)
{
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *button;
  GtkWidget *label;

  vbox = gtk_vbox_new(FALSE, 6);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);
  
  hbox = gtk_hbox_new(FALSE, 12);

  /* MTAs menu */
  label = gtk_label_new_with_mnemonic(_("_Sending Agent:"));
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

  ideditor->mtas_optmenu = gtk_option_menu_new();
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), ideditor->mtas_optmenu);
  gtk_widget_show(ideditor->mtas_optmenu);
  gtk_box_pack_start(GTK_BOX(hbox), ideditor->mtas_optmenu, TRUE, TRUE, 0);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);

  /* Props, Add, Remove */
  hbox = gtk_hbox_new(TRUE, 12);

  ideditor->mtas_props = gtk_button_new_from_stock(GTK_STOCK_PROPERTIES);
  gtk_widget_show(ideditor->mtas_props);
  gtk_box_pack_start(GTK_BOX(hbox), ideditor->mtas_props, TRUE, TRUE, 0);

  button = gtk_button_new_from_stock(GTK_STOCK_ADD);
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);

  ideditor->mtas_remove = gtk_button_new_from_stock(GTK_STOCK_REMOVE);
  gtk_widget_show(ideditor->mtas_remove);
  gtk_box_pack_start(GTK_BOX(hbox), ideditor->mtas_remove, TRUE, TRUE, 0);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);

  ideditor->mtaeditor = a_GtkMTAEditor_new(ideditor->widget);
  a_GtkMTAEditor_set_cancel_callback(ideditor->mtaeditor,
				     GtkIdEditor_mta_cancel_cb,
				     ideditor);
  g_signal_connect(G_OBJECT(ideditor->mtas_optmenu), "changed",
		   G_CALLBACK(GtkIdEditor_mta_changed_cb),
		   ideditor);

  g_signal_connect(G_OBJECT(ideditor->mtas_props), "clicked",
		   G_CALLBACK(GtkIdEditor_mta_edit_cb),
		   ideditor);

  g_signal_connect(G_OBJECT(button), "clicked",
		   G_CALLBACK(GtkIdEditor_mta_add_cb),
		   ideditor);

  g_signal_connect(G_OBJECT(ideditor->mtas_remove), "clicked",
		   G_CALLBACK(GtkIdEditor_mta_remove_cb),
		   ideditor);

  label = gtk_label_new_with_mnemonic(_("Sending _Agent"));
  gtk_widget_show(label);
  
  gtk_notebook_append_page(GTK_NOTEBOOK(ideditor->notebook), vbox, label);
  gtk_widget_show(vbox);
}

static void
GtkIdEditor_add_defaults_page(XePGtkIdEditor *ideditor)
{
  GtkWidget *label;
  GtkWidget *table;
  int i = 0;

  table = gtk_table_new(11, 2, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 12);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);

  ideditor->def_drafts =
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Drafts:"), NULL, 
			       NULL, 0, i);
  g_object_set_data(G_OBJECT(ideditor->def_drafts), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_DRAFTS_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->def_drafts), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);

  i++;

  ideditor->def_queue =
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Queue:"), NULL, 
			       NULL, 0, i);
  g_object_set_data(G_OBJECT(ideditor->def_queue), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_QUEUE_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->def_queue), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);

  i++;

  ideditor->def_sent =
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Sent:"), NULL,
			       NULL, 0, i);
  g_object_set_data(G_OBJECT(ideditor->def_sent), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_SENT_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->def_sent), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);


  i++;

  ideditor->def_templates = 
    a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Templates:"), NULL,
			       NULL, 0, i);
  g_object_set_data(G_OBJECT(ideditor->def_templates), "mask",
		    GUINT_TO_POINTER(XEP_IDEDITOR_TEMPLATES_CHANGED));
  g_signal_connect(G_OBJECT(ideditor->def_templates), "changed",
		   G_CALLBACK(GtkIdEditor_entry_changed), ideditor);

  /* Signal callbacks when entry is changed */

  label = gtk_label_new_with_mnemonic(_("_Defaults"));
  gtk_widget_show(label);
  
  gtk_notebook_append_page(GTK_NOTEBOOK(ideditor->notebook), table, label);
  gtk_widget_show(table);
}


static gboolean
GtkIdEditor_delete_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) data;
  GtkIdEditor_hide(ideditor);
  return TRUE;
}

static void
GtkIdEditor_destroy_cb(GObject *adialog, gpointer data)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) data;

  GtkIdEditor_identities_free_copy(ideditor);
  a_UI_id_editor_unlock();
  free(ideditor);
}

void
a_GtkIdEditor_show(XePUIMain *uimain)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) uimain->ideditor;

  if (a_UI_id_editor_is_locked())
    {
      a_GtkUtils_dialog_alert(_("An identity editor is already opened."), 
			      uimain->toplevel,
			      XETPAN_INFO_MESSAGE,
			      NULL,
			      NULL,
			      XEP_GTKUTILS_OK,
			      TRUE, TRUE, NULL, NULL);
      return;
    }

  if (ideditor == NULL)
    {
      if (GtkIdEditor_create(uimain) != 0)
	return;
      ideditor = (XePGtkIdEditor *) uimain->ideditor;
    }

  if (GtkIdEditor_identities_new_copy(ideditor) != 0)
    return;
  
  gtk_notebook_set_current_page(GTK_NOTEBOOK(ideditor->notebook), 
				XEP_IDEDITOR_IDENTITY);
  GtkIdEditor_init(ideditor, NULL);

  gtk_widget_show(GTK_WIDGET(ideditor->widget));
  a_UI_id_editor_lock();
}

static void
GtkIdEditor_hide(XePGtkIdEditor *ideditor)
{
  if (ideditor->opened_editors)
    {
      a_GtkUtils_dialog_alert(_("Identity Editor Busy"), ideditor->widget,
			      XETPAN_ERROR_MESSAGE,
			      _("Cannot close identity editor until all "
				"its windows have been closed."),
			      _("Please finish editing the signature, the "
				"mailbox or the sending agent "
				"before closing the identity editor."),
			      XEP_GTKUTILS_OK,
			      TRUE, /* modal */
			      TRUE, /* async */
			      NULL, NULL);
    }
  else
    {
      a_UI_id_editor_unlock();
      gtk_widget_hide(GTK_WIDGET(ideditor->widget));
      gtk_widget_set_sensitive(ideditor->widget, TRUE);
    }
}

static int
GtkIdEditor_identities_new_copy(XePGtkIdEditor *ideditor)
{
  clist *identities;
  clist *identity_list;
  clistiter *iter;
  XePUtilsTriplet *triplet;
  XePIdentity *identity;
  XePIdentity *identity_copy;
  int r = 0;

  A_LOGGING_TRACE;

  identities = a_Identity_list_get();

  identity_list = clist_new();
  if (identity_list == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      return -1;
    }

  for (iter = clist_begin(identities); iter && r == 0; iter = iter->next)
    {
      identity = (XePIdentity *) iter->data;
      identity_copy = a_Identity_new_copy(identity);
      if (identity_copy)
	{
	  triplet = x_alloc(1, XePUtilsTriplet);
	  if (triplet)
	    {
	      ID_TRIPLET_COPY(triplet) = identity_copy;
	      ID_TRIPLET_ORIG(triplet) = identity;
	      ID_TRIPLET_FLAGS(triplet) =  XEP_IDEDITOR_EXPANDED;
	      if (clist_append(identity_list, triplet))
		{
		  A_LOGGING_ERROR("malloc(): %s", strerror(errno));
		  a_Identity_free_copy(identity_copy);
		  free(triplet);
		  r = -1;
		}
	    }
	  else
	    {
	      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
	      a_Identity_free_copy(identity_copy);
	      r = -1;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  
  ideditor->identity_list = identity_list;

  if (r == -1)
    {
      GtkIdEditor_identities_free_copy(ideditor);
    }

  return r;
}

static void
GtkIdEditor_init(XePGtkIdEditor *ideditor, XePUtilsTriplet *show_triplet)
{
  GtkWidget *menu;
  GtkWidget *menuitem;
  XePUtilsTriplet *triplet;
  XePUtilsTriplet *current_triplet;
  XePIdentity *identity_copy;
  clistiter *iter;
  gboolean is_default = TRUE;
  gint i, pos;
  current_triplet = show_triplet;

  menu = gtk_menu_new();
  pos = i = 0;
  for (iter = clist_begin(ideditor->identity_list); iter; iter = iter->next)
    {
      triplet = (XePUtilsTriplet *) iter->data;
      identity_copy = ID_TRIPLET_COPY(triplet);
      if (identity_copy)
	{
	  menuitem = gtk_menu_item_new_with_label(identity_copy->id);
	  if (is_default)
	    {
	      /* Default id is the first one */
	      if (current_triplet == NULL) 
		current_triplet = triplet;
	      is_default = FALSE;
	    }
	  if (current_triplet == triplet)
	    {
	      pos = i;
	      ideditor->current_id_label = 
		gtk_bin_get_child(GTK_BIN(menuitem));
	    }
	  gtk_widget_show(menuitem);
	  g_object_set_data(G_OBJECT(menuitem), "identity", triplet);
	  g_signal_connect(G_OBJECT(menuitem), "activate",
			   G_CALLBACK(GtkIdEditor_identity_activate_cb), ideditor);
	  gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
	  i++;
	}
    }
  ideditor->current_triplet = current_triplet;
	 
  gtk_option_menu_remove_menu(GTK_OPTION_MENU(ideditor->id_option_menu));
  gtk_option_menu_set_menu(GTK_OPTION_MENU(ideditor->id_option_menu), menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(ideditor->id_option_menu), pos);
  gtk_widget_show(menu);

  GtkIdEditor_update_pages(ideditor);
}

static void
GtkIdEditor_identities_free_copy(XePGtkIdEditor *ideditor)
{
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;
  clistiter *iter;

  A_LOGGING_TRACE;

  if (ideditor->identity_list == NULL)
    return;
    
  for (iter = clist_begin(ideditor->identity_list); iter; iter = iter->next)
    {
      triplet = (XePUtilsTriplet *) iter->data;
      identity_copy = ID_TRIPLET_COPY(triplet);
      if (identity_copy)
	{
	  a_Identity_free_copy(identity_copy);
	}
      free(triplet);
    }
  clist_free(ideditor->identity_list);

  ideditor->identity_list = NULL;
}

static void
GtkIdEditor_update_pages(XePGtkIdEditor *ideditor)
{
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy = NULL;
  GtkNotebook *notebook = GTK_NOTEBOOK(ideditor->notebook);
  gboolean sensitive;

  triplet = ideditor->current_triplet;
  
  if (triplet)
    {
      identity_copy = ID_TRIPLET_COPY(triplet);
      
      /* XEP_IDEDITOR_IDENTITY: */
      a_GtkUtils_entry_set_text(ideditor->id_name, 
				a_Identity_get_name(identity_copy));
      a_GtkUtils_entry_set_text(ideditor->id_email, 
				a_Identity_get_email(identity_copy));
      a_GtkUtils_entry_set_text(ideditor->id_org, 
				a_Identity_get_organization(identity_copy));
      a_GtkUtils_entry_set_text(ideditor->id_reply_to, 
				a_Identity_get_reply_to(identity_copy));
      
      /* XEP_IDEDITOR_SIGNATURES: */
      a_GtkUtils_tree_set_list(ideditor->sigtree, 
			       a_Identity_get_signatures(identity_copy));
      a_GtkUtils_tree_refresh(ideditor->sigtree);
      /* XEP_IDEDITOR_MAILBOXES: */
      a_GtkUtils_tree_refresh(ideditor->mbtree);
      /* XEP_IDEDITOR_MTAS: */
      GtkIdEditor_mtas_update_menu(ideditor, 
				   a_Identity_get_mta_ref(identity_copy));
      /* XEP_IDEDITOR_DEFAULTS: */
      a_GtkUtils_entry_set_text(ideditor->def_drafts,
				a_Identity_get_draft_ref(identity_copy));
      a_GtkUtils_entry_set_text(ideditor->def_queue, 
				a_Identity_get_queue_ref(identity_copy));
      a_GtkUtils_entry_set_text(ideditor->def_sent, 
				a_Identity_get_sent_ref(identity_copy));
      a_GtkUtils_entry_set_text(ideditor->def_templates, 
				a_Identity_get_templates_ref(identity_copy));

      gtk_widget_set_sensitive(ideditor->id_option_menu, TRUE);
      gtk_widget_set_sensitive(ideditor->id_delete, TRUE);
      {
	clistiter *iter = clist_begin(ideditor->identity_list);
	gtk_widget_set_sensitive(ideditor->id_default,
				 iter->data != triplet);
      }	  
      sensitive = TRUE;
    }
  else
    {
      gtk_widget_set_sensitive(ideditor->id_option_menu, FALSE);
      gtk_widget_set_sensitive(ideditor->id_delete, FALSE);
      sensitive = FALSE;
    }

  gtk_widget_set_sensitive(gtk_notebook_get_nth_page(notebook, 
						     XEP_IDEDITOR_DEFAULTS),
			   sensitive);
  gtk_widget_set_sensitive(gtk_notebook_get_nth_page(notebook, 
					    XEP_IDEDITOR_MTAS),
			   sensitive);
  gtk_widget_set_sensitive(gtk_notebook_get_nth_page(notebook, 
					    XEP_IDEDITOR_MAILBOXES),
			   sensitive);      
  gtk_widget_set_sensitive(gtk_notebook_get_nth_page(notebook, 
						     XEP_IDEDITOR_SIGNATURES),
			   sensitive);
  gtk_widget_set_sensitive(gtk_notebook_get_nth_page(notebook, 
						     XEP_IDEDITOR_IDENTITY),
			   sensitive);
}

static void
GtkIdEditor_set_opened_editor(XePGtkIdEditor *ideditor, 
			      XePGtkIdEditorPage page,
			      gboolean open)
{
  GtkWidget *widget;

  if (open)
    ideditor->opened_editors++;
  else
    ideditor->opened_editors--;

  gtk_widget_set_sensitive(ideditor->buttons, !open);
  gtk_widget_set_sensitive(ideditor->id_new, !open);
  gtk_widget_set_sensitive(ideditor->id_delete, !open);
  gtk_widget_set_sensitive(ideditor->id_option_menu, !open);

  widget = gtk_notebook_get_nth_page(GTK_NOTEBOOK(ideditor->notebook), page);
  gtk_widget_set_sensitive(widget, !open);
}

/* Coordinates */

static void
GtkIdEditor_entry_changed(GtkEditable *editable, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  unsigned int mask;
  XePUtilsTriplet *triplet;
  XePIdentity *identity;
  unsigned int *flags;
  GtkWidget *entry;
  const char *orig;
  const char *text;
  unsigned int before;
  int changed;

  mask = (unsigned int) g_object_get_data(G_OBJECT(editable), "mask");

  triplet = ideditor->current_triplet;

  identity =  ID_TRIPLET_ORIG(triplet);
  /* If this is a newly added identity, the associated tree is
   * already marked as changed
   */
  if (identity == NULL)
    return;
    
  switch (mask)
    {
    case XEP_IDEDITOR_NAME_CHANGED:
      orig = a_Identity_get_name(identity);
      entry = ideditor->id_name;
      break;
    case XEP_IDEDITOR_EMAIL_CHANGED:
      orig = a_Identity_get_email(identity);
      entry = ideditor->id_email;
      break;
    case XEP_IDEDITOR_ORG_CHANGED:
      orig = a_Identity_get_organization(identity);
      entry = ideditor->id_org;
      break;
    case XEP_IDEDITOR_REPLY_TO_CHANGED:
      orig = a_Identity_get_reply_to(identity);
      entry = ideditor->id_reply_to;
      break;
    case XEP_IDEDITOR_DRAFTS_CHANGED:
      orig = a_Identity_get_draft_ref(identity);
      entry = ideditor->def_drafts;
      break;
    case XEP_IDEDITOR_QUEUE_CHANGED:
      orig = a_Identity_get_queue_ref(identity);
      entry = ideditor->def_queue;
      break;
    case XEP_IDEDITOR_SENT_CHANGED:
      orig = a_Identity_get_sent_ref(identity);
      entry = ideditor->def_sent;
      break;
    case XEP_IDEDITOR_TEMPLATES_CHANGED:
      orig = a_Identity_get_templates_ref(identity);
      entry = ideditor->def_templates;
      break;
    default:
      return;
    }

  text = gtk_entry_get_text(GTK_ENTRY(entry));
  if (orig == NULL)
    changed = (text[0] != 0x00);
  else
    changed = strcmp(orig, text);

  flags  = (int *) (&(triplet->data3));
  before = *flags & mask;
  if (!before && changed)
    {
      *flags = *flags | mask;
    }

  if (before &&  !changed)
    {
      *flags = *flags & ~mask;
    }

  if (mask & (XEP_IDEDITOR_NAME_CHANGED | XEP_IDEDITOR_EMAIL_CHANGED))
    {
      const char *name;
      const char *email;
      if (mask & XEP_IDEDITOR_NAME_CHANGED)
	{
	  name = text;
	  email = gtk_entry_get_text(GTK_ENTRY(ideditor->id_email));
	}
      else
	{
	  email = text;
	  name =  gtk_entry_get_text(GTK_ENTRY(ideditor->id_name));
	}
      gtk_label_set_text(GTK_LABEL(ideditor->current_id_label), 
			 (name ? name : ""));
    }
}

/* Identity Coordinates */

static void GtkIdEditor_identity_activate_cb(GtkMenuItem *menuitem, 
					     gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  A_LOGGING_TRACE;

  GtkIdEditor_apply_changes_to_copy(ideditor);

  ideditor->current_id_label = gtk_bin_get_child(GTK_BIN(menuitem));
  ideditor->current_triplet = g_object_get_data(G_OBJECT(menuitem),
						"identity");
  GtkIdEditor_update_pages(ideditor);
}


static void
GtkIdEditor_identity_add_cb(GtkButton* button, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *new_triplet;
  XePIdentity *new_identity;

  GtkIdEditor_apply_changes_to_copy(ideditor);

  new_identity = a_Identity_new();
  if (new_identity == NULL)
    return;

  a_Identity_set_id(new_identity, _("New Identity"));
  a_Identity_set_name(new_identity, _("New Identity"));
  a_Identity_set_email(new_identity, "");

  new_triplet = x_alloc(1, XePUtilsTriplet);
  if (new_triplet == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      a_Identity_free(new_identity);
      return;
    }
  ID_TRIPLET_COPY(new_triplet) = new_identity;
  ID_TRIPLET_ORIG(new_triplet) = NULL;
  ID_TRIPLET_FLAGS(new_triplet) = XEP_IDEDITOR_IDENTITY_CHANGED | 
    XEP_IDEDITOR_SIGNATURES_CHANGED |
    XEP_IDEDITOR_MAILBOXES_CHANGED  |
    XEP_IDEDITOR_MTAS_CHANGED       |
    XEP_IDEDITOR_DEFAULTS_CHANGED   |
    XEP_IDEDITOR_NEW                |
    XEP_IDEDITOR_SELECT             |
    XEP_IDEDITOR_EXPANDED;

  clist_append(ideditor->identity_list, new_triplet);
  GtkIdEditor_init(ideditor, new_triplet);
}

static void
GtkIdEditor_identity_remove_cb(GtkButton *button, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;
  
  triplet = ideditor->current_triplet;

  identity_copy = ID_TRIPLET_COPY(triplet);
  a_Identity_free_copy(identity_copy);
  ID_TRIPLET_COPY(triplet) = NULL;

  GtkIdEditor_init(ideditor, NULL);
}

static void
GtkIdEditor_identity_set_default_cb(GtkButton *button, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  clistiter *iter;

  triplet = ideditor->current_triplet;
			
  for (iter = clist_begin(ideditor->identity_list); 
       iter && iter->data != triplet; 
       iter = iter->next)
    {
      /* empty on purpose */
    }

  if (iter) 
    {
      clist_delete(ideditor->identity_list, iter);
      clist_prepend(ideditor->identity_list, triplet);
    }

  GtkIdEditor_init(ideditor, triplet);
}



static void
GtkIdEditor_identity_set_changed(XePGtkIdEditor *ideditor, 
				 unsigned int set, unsigned int unset)
{
  XePUtilsTriplet *triplet;
  unsigned int new_flags;

  triplet = ideditor->current_triplet;

  new_flags = (ID_TRIPLET_FLAGS(triplet) & ~unset) | set;

  if (new_flags != ID_TRIPLET_FLAGS(triplet))
    {
      ID_TRIPLET_FLAGS(triplet) = new_flags;
    }
}

/* Signatures */

static void *
GtkIdEditor_sig_add_cb(XePGtkUtilsTree *sigtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;
  gchar *title;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_SIGNATURES, TRUE);

  triplet = ideditor->current_triplet;
    
  identity_copy = ID_TRIPLET_COPY(triplet);

  if (identity_copy->name || identity_copy->email)
    {
      title = g_strdup_printf(_("New signature file for %s"), 
			      (identity_copy->name ? 
			       identity_copy->name : 
			       identity_copy->email));
    }
  else
    {
      title = g_strdup_printf(_("New signature file"));
    }
			    
  a_GtkUtils_dialog_get_string(title,
			       ideditor->widget,
			       _("Signature file:"),
			       _("Enter the path of your signature file. "
				 "If the signature should be the output of "
				 "a shell command, enter the command and "
				 "add the vertical bar \"|\" as the last "
				 "character."),
			       NULL,
			       XEP_GTKUTILS_OK | XEP_GTKUTILS_CANCEL |
			       XEP_GTKUTILS_FILESEL,
			       FALSE,
			       GtkIdEditor_sig_add_real_cb,
			       ideditor);
  g_free(title);
  return NULL;
}

static void
GtkIdEditor_sig_add_real_cb(void *aideditor, int response, gchar *signature)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_SIGNATURES, FALSE);

  if (response == GTK_RESPONSE_CANCEL || 
      signature == NULL || 
      signature[0] == 0x00)
    {
      return;
    }

  a_GtkUtils_tree_finalize_add(ideditor->sigtree, signature);
}

static void *
GtkIdEditor_sig_remove_cb(XePGtkUtilsTree *sigtree, void *aideditor)
{
  clistiter *iter = NULL;
  char *signature;

  a_GtkUtils_tree_get_selected_data(sigtree, NULL, NULL, NULL,
				    (void **) iter, NULL);

  signature = iter->data;
  x_free(signature);
  return NULL;
}

static void *
GtkIdEditor_sig_changed_cb(XePGtkUtilsTree *sigtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_identity_set_changed(ideditor, XEP_IDEDITOR_SIGNATURES_CHANGED,
				   0);

  return NULL;
}

static void *
GtkIdEditor_sig_refresh_cb(XePGtkUtilsTree *sigtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  GtkTreeStore *treestore;
  clistiter *iter;
  const char *signature;
  GtkTreeIter treeiter;
  XePIdentity *identity_copy;
  
  triplet = ideditor->current_triplet;

  treestore = a_GtkUtils_tree_get_tree_store(sigtree);

  identity_copy = ID_TRIPLET_COPY(triplet);

  for (iter = clist_begin(identity_copy->signatures); iter; iter = iter->next)
    {
      signature = (const char *) iter->data;
      gtk_tree_store_append(treestore, &treeiter, NULL);
      gtk_tree_store_set(treestore, &treeiter, 
			 TREE_DESCRIPTION_COL, signature,
			 TREE_TYPE_COL, 0,
			 TREE_WEIGHT_COL, PANGO_WEIGHT_NORMAL,
			 TREE_POINTER_COL, iter,
			 TREE_POINTER2_COL, NULL,
			 -1);
    }
  return NULL;
}

/* Mailboxes */

static void
GtkIdEditor_mailboxes_cancel_cb(XePGtkMbEditor *mbeditor, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MAILBOXES, FALSE);
}

static void *
GtkIdEditor_mailboxes_add_cb(XePGtkUtilsTree *mbtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MAILBOXES, TRUE);

  a_GtkMbEditor_set_ok_callback(ideditor->mbeditor, 
				GtkIdEditor_mailboxes_add_done_cb, 
				ideditor);
  a_GtkMbEditor_set_mailbox(ideditor->mbeditor, NULL);
  a_GtkMbEditor_show(ideditor->mbeditor);

  return NULL;
}

static void
GtkIdEditor_mailboxes_add_done_cb(XePGtkMbEditor *mbeditor, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;
  clistiter *iter;
  XePMailbox *mailbox;

  triplet = ideditor->current_triplet;

  identity_copy = ID_TRIPLET_COPY(triplet);

  a_GtkUtils_tree_get_selected_data(ideditor->mbtree, NULL, NULL, NULL,
				    (void **)&iter, NULL);

  mailbox = a_GtkMbEditor_get_mailbox(mbeditor);

  clist_insert_after(identity_copy->mailboxes, iter, mailbox);

  a_GtkUtils_tree_finalize_add(ideditor->mbtree, mailbox);

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MAILBOXES, FALSE);
}

static void *
GtkIdEditor_mailboxes_remove_cb(XePGtkUtilsTree *mbtree, void *aideditor)
{
  XePMailbox *mailbox;
  clistiter *iter = NULL;

  a_GtkUtils_tree_get_selected_data(mbtree, NULL, NULL, NULL,
				    (void **)&iter, NULL);
  
  mailbox = (XePMailbox *) iter->data;

  mailbox->uid = -(mailbox->uid); /* Mark to be deleted */
  a_GtkUtils_tree_refresh(mbtree);
  return NULL;
}

static void *
GtkIdEditor_mailboxes_refresh_cb(XePGtkUtilsTree *mbtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  GtkTreeStore *treestore;
  clistiter *iter;
  XePMailbox *mailbox;
  GtkTreeIter treeiter;
  XePIdentity *identity_copy;
  
  triplet = ideditor->current_triplet;

  treestore = a_GtkUtils_tree_get_tree_store(mbtree);

  identity_copy = ID_TRIPLET_COPY(triplet);

  for (iter = clist_begin(identity_copy->mailboxes); iter; iter = iter->next)
    {
      mailbox = (XePMailbox *) iter->data;
      if (mailbox->uid < 0)
	continue;
      gtk_tree_store_append(treestore, &treeiter, NULL);
      gtk_tree_store_set(treestore, &treeiter, 
			 TREE_DESCRIPTION_COL, mailbox->id,
			 TREE_TYPE_COL, 0,
			 TREE_WEIGHT_COL, PANGO_WEIGHT_NORMAL,
			 TREE_POINTER_COL, iter,
			 TREE_POINTER2_COL, NULL,
			 -1);
    }
  return NULL;

}

static void *
GtkIdEditor_mailboxes_edit_cb(XePGtkUtilsTree *mbtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  clistiter *iter;
  XePMailbox *mailbox;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MAILBOXES, TRUE);

  a_GtkUtils_tree_get_selected_data(mbtree, NULL, NULL, NULL,
				    (void **)&iter, NULL);
  mailbox = (XePMailbox *) iter->data;

  if (ideditor->mbeditor == NULL)
    {
      ideditor->mbeditor = a_GtkMbEditor_new(ideditor->widget);
    }

  a_GtkMbEditor_set_ok_callback(ideditor->mbeditor, 
				GtkIdEditor_mailboxes_edit_done_cb,
				ideditor);
  a_GtkMbEditor_set_mailbox(ideditor->mbeditor, mailbox);
  a_GtkMbEditor_show(ideditor->mbeditor);

  return NULL;
}

static void
GtkIdEditor_mailboxes_edit_done_cb(XePGtkMbEditor *mbeditor, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePMailbox *mailbox;

  mailbox = a_GtkMbEditor_get_mailbox(mbeditor);
  GtkIdEditor_identity_set_changed(ideditor, XEP_IDEDITOR_MAILBOXES_CHANGED,
				   0);

  a_GtkUtils_tree_finalize_edit(ideditor->mbtree, mailbox);

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MAILBOXES, FALSE);
}

static void *
GtkIdEditor_mailboxes_changed_cb(XePGtkUtilsTree *mbtree, void *aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_identity_set_changed(ideditor, XEP_IDEDITOR_MAILBOXES_CHANGED,
				   0);

  return NULL;
}
/* MTAs */

static void
GtkIdEditor_mta_changed_cb(GtkOptionMenu *optmenu, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  const char *def_mta_id;
  XePUtilsTriplet *triplet = ideditor->current_triplet;
  XePIdentity *identity = ID_TRIPLET_ORIG(triplet);

  def_mta_id = GtkIdEditor_mtas_get_selected(ideditor);

  if (identity)
    {
      const char *old_mta_id = a_Identity_get_mta_ref(identity);
      if (!old_mta_id || strcmp(old_mta_id, def_mta_id))
	ID_TRIPLET_FLAGS(triplet) |= XEP_IDEDITOR_MTAS_CHANGED;
    }
  else
    {
      ID_TRIPLET_FLAGS(triplet) |= XEP_IDEDITOR_MTAS_CHANGED;
    }
}

static const char *
GtkIdEditor_mtas_get_selected(XePGtkIdEditor *ideditor)
{
  clist *mtas;
  XePMTA *mta;
  guint selected;

  selected = 
    gtk_option_menu_get_history(GTK_OPTION_MENU(ideditor->mtas_optmenu));

  mtas = a_MTA_list_get();

  mta = (XePMTA *) clist_nth_data(mtas, selected);

  return (mta ? a_MTA_get_id(mta) : NULL);
}

static void
GtkIdEditor_mta_cancel_cb(XePGtkMTAEditor *mtaeditor, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MTAS, FALSE);
}

static void
GtkIdEditor_mta_add_cb(GtkButton *button, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;

  triplet = ideditor->current_triplet;

  identity_copy = ID_TRIPLET_COPY(triplet);

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MTAS, TRUE);

  a_GtkMTAEditor_set_ok_callback(ideditor->mtaeditor,
				 GtkIdEditor_mta_add_done_cb, 
				 ideditor);
  a_GtkMTAEditor_set_mta(ideditor->mtaeditor, NULL);
  a_GtkMTAEditor_show(ideditor->mtaeditor);
}

static void
GtkIdEditor_mta_add_done_cb(XePGtkMTAEditor *mtaeditor, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePMTA *mta;
  clistiter *iter;
  clist *mtas;
  guint selected;

  mta = a_GtkMTAEditor_get_mta(ideditor->mtaeditor);

  mtas = a_MTA_list_get();
  selected = 
    gtk_option_menu_get_history(GTK_OPTION_MENU(ideditor->mtas_optmenu));

  iter = clist_nth(mtas, selected);
  clist_insert_after(mtas, iter, mta);
  GtkIdEditor_mtas_update_menu(ideditor, mta->id);
}

static void
GtkIdEditor_mta_edit_cb(GtkButton *button, gpointer aideditor)

{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;
  XePMTA *mta;
  clist *mtas;
  guint selected;

  triplet = ideditor->current_triplet;

  identity_copy = ID_TRIPLET_COPY(triplet);

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MTAS, TRUE);

  selected = 
    gtk_option_menu_get_history(GTK_OPTION_MENU(ideditor->mtas_optmenu));

  mtas = a_MTA_list_get();

  mta = clist_nth_data(mtas, selected);

  a_GtkMTAEditor_set_ok_callback(ideditor->mtaeditor, 
				 GtkIdEditor_mta_edit_done_cb, ideditor);
  a_GtkMTAEditor_set_mta(ideditor->mtaeditor, mta);
  a_GtkMTAEditor_show(ideditor->mtaeditor);
}

static void
GtkIdEditor_mta_edit_done_cb(XePGtkMTAEditor *mtaeditor, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;

  GtkIdEditor_set_opened_editor(ideditor, XEP_IDEDITOR_MTAS, FALSE);
}

static void
GtkIdEditor_mta_remove_cb(GtkButton *button, gpointer aideditor)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) aideditor;
  XePUtilsTriplet *triplet;
  XePIdentity *identity, *identity_copy;
  XePMTA *mta;
  const char *mta_id;
  clist *mtas;
  clistiter *iter;
  guint selected;
  GSList *ids = NULL;
  GSList *list;
  size_t len = 0;

  triplet = ideditor->current_triplet;

  identity_copy = ID_TRIPLET_COPY(triplet);

  selected = 
    gtk_option_menu_get_history(GTK_OPTION_MENU(ideditor->mtas_optmenu));

  mtas = a_MTA_list_get();

  mta = clist_nth_data(mtas, selected);
  mta_id = a_MTA_get_id(mta);

  /* List all other identities using this MTA */
  for (iter = clist_begin(ideditor->identity_list); iter; iter = iter->next)
    {
      const char *id;
      triplet = (XePUtilsTriplet *) iter->data;
      identity = ID_TRIPLET_COPY(triplet);
      id = a_Identity_get_mta_ref(identity);

      if (identity != identity_copy && id && strcmp(id, mta_id) == 0)
	{
	  id = a_Identity_get_id(identity);
	  ids = g_slist_append(ids, (char *) id);
	  len += strlen(id) + 1;
	}
    }

  /* If there are other identities, abort MTA removal */
  if (ids) 
    {
      const char *msg = _("There are other identities that are currently "
			  "using this sending agent.\n"
			  "Other identities using this sending agent are:\n");
			  
      char *text;
      char *p;
      text = x_alloc(strlen(msg) + len + 1, char);

      strcpy(text, msg);
      p = text + strlen(msg);
      for (list = ids; list; list = list->next)
	{
	  strcpy(p, (char *)(list->data));
	  p += strlen((char *)(list->data));
	  p[0] = '\n';
	  p++;
	}
      g_slist_free(ids);
      p[0] = 0x00;
      a_GtkUtils_dialog_alert(_("Removing a Sending Agent"),
			      ideditor->widget,
			      XETPAN_WARNING_MESSAGE,
			      text,
			      _("To remove this sending agent, first make "
				"sure that no other identity but this one is u"
				"sing it."),
			      XEP_GTKUTILS_OK,
			      TRUE,
			      TRUE,
			      NULL, NULL);

      return;
    }

  a_MTA_free(mta);
  iter = clist_nth(mtas, selected);
  clist_delete(mtas, iter);

  GtkIdEditor_mtas_update_menu(ideditor, NULL);
}

static void 
GtkIdEditor_mtas_update_menu(XePGtkIdEditor *ideditor, const char *def_mta)
{
  GtkWidget *menu;
  GtkWidget *menuitem;
  clistiter *iter;
  clist *mtas;
  XePMTA *mta;
  gboolean has_mtas;
  guint i;
  guint selected = 0;
  const char *mta_id;

  gtk_option_menu_remove_menu(GTK_OPTION_MENU(ideditor->mtas_optmenu));

  menu = gtk_menu_new();
  
  mtas = a_MTA_list_get();

  for (iter = clist_begin(mtas), i = 0; iter; iter = iter->next, i++)
    {
      mta = (XePMTA *) iter->data;
      mta_id = a_MTA_get_id(mta);
      if (def_mta && mta_id && strcmp(mta_id, def_mta) == 0)
	selected = i;
      menuitem = gtk_menu_item_new_with_label(mta_id);
      gtk_widget_show(menuitem);
      gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
    }

  has_mtas = clist_count(mtas);
  gtk_widget_set_sensitive(ideditor->mtas_props, has_mtas);
  gtk_widget_set_sensitive(ideditor->mtas_remove, has_mtas);

  gtk_option_menu_set_menu(GTK_OPTION_MENU(ideditor->mtas_optmenu), menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(ideditor->mtas_optmenu), 
			      selected);
}

/* Cancel and OK actions */

static void
GtkIdEditor_cancel_cb(GtkButton *button, gpointer data)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) data;
  GtkIdEditor_identities_free_copy(ideditor);
  GtkIdEditor_hide(ideditor);
}

static void
GtkIdEditor_ok_cb(GtkButton *button, gpointer data)
{
  XePGtkIdEditor *ideditor = (XePGtkIdEditor *) data;
  clist *identities;
  clistiter *iter;
  XePUtilsTriplet *triplet;
  XePIdentity *identity_copy;
  XePIdentity *identity;

  A_LOGGING_TRACE;
  
  gtk_widget_set_sensitive(ideditor->widget, FALSE);

  /* First, transfert UI fields to the identity_copy */
  GtkIdEditor_apply_changes_to_copy(ideditor);
  
  identities = a_Identity_list_get();

  /* TODO: apply changes to the identities that are not used, 
   * and notify the user for the identities that were not updated
   * because they were busy
   */
  for (iter = clist_begin(identities); iter; iter = iter->next)
    {
      identity = (XePIdentity *) iter->data;
      if (a_Identity_is_busy(identity))
	{
	  char *msg;
	  msg = g_strdup_printf(_("Changes to identity %s cannot be applied because "
				  "it is currently being used.  Try later."),
				identity->id);
	  a_GtkUtils_dialog_alert(_("Operation not permitted"),
				  ideditor->widget, XETPAN_INFO_MESSAGE,
				  msg,
				  _("You can check what operations are "
				    "currently running by opening the task "
				    "manager (joke)."),
				  XEP_GTKUTILS_OK, TRUE, TRUE, NULL, NULL);
	  gtk_widget_set_sensitive(ideditor->widget, TRUE);
	  g_free(msg);
	  return;				  
	}
    }

  /* Empty the global identity list without freeing the identities 
   * themselves. They are not lost: they are in identity_list triplets
   */
  for (iter = clist_begin(identities); 
       iter;
       iter = clist_delete(identities, iter));

  /* Add one to the id_count to avoid calling finalize code before the 
   * loop has finished
   */
  ideditor->id_count = 1;

  for (iter = clist_begin(ideditor->identity_list); iter; iter = iter->next)
    {
      XePUtilsPair *changes;
      
      triplet = (XePUtilsTriplet *) iter->data;
      identity_copy = ID_TRIPLET_COPY(triplet);
      identity = ID_TRIPLET_ORIG(triplet);

      /* Some changes may occure asynchronously, so we store their data here 
       * They will be passed to the callback functions called when changes
       * finish
       */
      changes = x_alloc(1, XePUtilsPair);
      ID_CHANGES_IDEDITOR(changes) = ideditor;
      ID_CHANGES_TRIPLET(changes) = triplet;

      if (identity_copy)
	{
	  /* Create a new one (else change the existing one) */
	  if (identity == NULL)
	    {
	      /* Create new identity, 
	       */
	      identity = a_Identity_new();
	      ID_TRIPLET_ORIG(triplet) = identity;
	      ID_TRIPLET_FLAGS(triplet) = XEP_IDEDITOR_ALL_CHANGED;
	    }
	  
	  clist_append(identities, identity);

	  /* Transfer changes from the copy to the identity */
	  ideditor->id_count++;
	  GtkIdEditor_transfer_changes(changes);
	}
      else
	{
	  /* Removed identity: disconnect it before freeing it */
	  identity = ID_TRIPLET_ORIG(triplet);
	  if (identity)
	    {
	      ideditor->id_count++;
	      a_Identity_disconnect_all(identity,
					GtkIdEditor_identity_free_finalize,
					changes);
	    }
	}
    }
  /* GtkIdEditor_transfer_changes() and GtkIdEditor_identity_free_finalize()
   * may modify ideditor->id_count before they return if no asynchronous
   * operation is needed, that's why we added one to the id_count
   * now we can decrement it 
   */
  ideditor->id_count--;
  if (ideditor->id_count == 0)
    GtkIdEditor_ok_finalize(ideditor);
}

static void
GtkIdEditor_ok_finalize(XePGtkIdEditor *ideditor)
{
  A_LOGGING_TRACE;

  GtkIdEditor_identities_free_copy(ideditor);

  a_Folder_ref_relink_all();
  
  a_Foldersview_mailboxes_view_update();
  a_Foldersview_list_expunge();
  a_Conf_write();

  a_UI_foldersview_list_updated();
  a_UI_update_folder(NULL);
  
  GtkIdEditor_hide(ideditor);
}

static void
GtkIdEditor_transfer_changes(XePUtilsPair *changes)
{
  XePUtilsTriplet *triplet = ID_CHANGES_TRIPLET(changes);
  XePIdentity *identity = ID_TRIPLET_ORIG(triplet);
  XePIdentity *identity_copy = ID_TRIPLET_COPY(triplet);
  int flags = (ID_TRIPLET_FLAGS(triplet));
  int mailbox_changed = 0;

  A_LOGGING_TRACE;

  if (flags & XEP_IDEDITOR_IDENTITY_CHANGED)
    {
      a_Identity_set_id(identity, a_Identity_get_id(identity_copy));
      a_Identity_set_name(identity, a_Identity_get_name(identity_copy));
      a_Identity_set_email(identity, a_Identity_get_email(identity_copy));
      a_Identity_set_organization(identity,
				  a_Identity_get_organization(identity_copy));
      a_Identity_set_reply_to(identity, 
			      a_Identity_get_reply_to(identity_copy));
      
      flags = flags & ~XEP_IDEDITOR_IDENTITY_CHANGED;
    }

  if (flags & XEP_IDEDITOR_SIGNATURES_CHANGED)
    {
      clist *list;

      list = a_Identity_signatures_new_copy(a_Identity_get_signatures(identity_copy));
      a_Identity_set_signatures(identity, list);

      flags = flags & ~XEP_IDEDITOR_SIGNATURES_CHANGED;
    }

  if(flags & XEP_IDEDITOR_MAILBOXES_CHANGED)
    {
      mailbox_changed = 1;
      /* FIXME: we should test for return value */
      /* GtkIdEditor_transfer_changes_finalize_clone()
       * will call GtkIdEditor_identity_finalize() on completion
       */
      a_Mailbox_list_clone(a_Identity_get_mailboxes(identity_copy),
			   GtkIdEditor_transfer_changes_finalize_clone,
			   changes);
    }

  if (flags & XEP_IDEDITOR_MTAS_CHANGED)
    {
      a_Identity_set_mta_ref(identity, a_Identity_get_mta_ref(identity_copy));

      flags = flags & ~XEP_IDEDITOR_MTAS_CHANGED;
    }

  if (flags & XEP_IDEDITOR_DEFAULTS_CHANGED)
    {
      a_Identity_set_draft_ref(identity, 
			       a_Identity_get_draft_ref(identity_copy));
      a_Identity_set_queue_ref(identity, 
			       a_Identity_get_queue_ref(identity_copy));
      a_Identity_set_sent_ref(identity, 
			      a_Identity_get_sent_ref(identity_copy));
      a_Identity_set_templates_ref(identity, 
				   a_Identity_get_templates_ref(identity_copy));

      flags = flags & ~XEP_IDEDITOR_DEFAULTS_CHANGED;
    }

  ID_TRIPLET_FLAGS(triplet) = flags;

  /* if mailbox did not change we have to finalize here
   */
  if (mailbox_changed == 0)
    GtkIdEditor_identity_finalize(changes);
}

static void
GtkIdEditor_transfer_changes_finalize_clone(clist *mailboxes, void *achanges)
{
  XePUtilsPair *changes  = (XePUtilsPair *) achanges;
  XePUtilsTriplet *triplet = ID_CHANGES_TRIPLET(changes);
  XePIdentity *identity = ID_TRIPLET_ORIG(triplet);

  A_LOGGING_TRACE;

  ID_TRIPLET_FLAGS(triplet) &= ~XEP_IDEDITOR_MAILBOXES_CHANGED;
  
  free(identity->mailboxes);
  identity->mailboxes = mailboxes;

  GtkIdEditor_identity_finalize(changes);
}

static void
GtkIdEditor_identity_finalize(XePUtilsPair *changes)
{
  XePGtkIdEditor *ideditor = ID_CHANGES_IDEDITOR(changes);

  A_LOGGING_TRACE;

  free(changes);

  ideditor->id_count--;
  if (ideditor->id_count)
    {
      return;
    }

  GtkIdEditor_ok_finalize(ideditor);
}

static void
GtkIdEditor_identity_free_finalize(XePIdentity *identity, void *achanges)
{
  XePUtilsPair *changes = (XePUtilsPair *) achanges;
  XePGtkIdEditor *ideditor = ID_CHANGES_IDEDITOR(changes);

  A_LOGGING_TRACE;

  free(changes);

  a_Identity_free(identity);

  ideditor->id_count--;
  if (ideditor->id_count)
    {
      return;
    }

  GtkIdEditor_ok_finalize(ideditor);
}

static void
GtkIdEditor_apply_changes_to_copy(XePGtkIdEditor *ideditor)
{
  XePIdentity *identity_copy;
  XePUtilsTriplet *triplet = ideditor->current_triplet;
  unsigned int flags;

  A_LOGGING_TRACE;

  if (triplet == NULL)
    {
      return;
    }

  identity_copy = ID_TRIPLET_COPY(triplet);
  flags = ID_TRIPLET_FLAGS(triplet);

  /* XEP_IDEDITOR_IDENTITY */

  /* For the moment, identity ID is the same as the display name */
  if (flags & XEP_IDEDITOR_NAME_CHANGED)
    a_Identity_set_id(identity_copy, 
		      gtk_entry_get_text(GTK_ENTRY(ideditor->id_name)));
  if (flags & XEP_IDEDITOR_NAME_CHANGED)
    a_Identity_set_name(identity_copy, 
			gtk_entry_get_text(GTK_ENTRY(ideditor->id_name)));
  if (flags & XEP_IDEDITOR_EMAIL_CHANGED)
    a_Identity_set_email(identity_copy, 
			 gtk_entry_get_text(GTK_ENTRY(ideditor->id_email)));
  if (flags & XEP_IDEDITOR_ORG_CHANGED)
    a_Identity_set_organization(identity_copy,
				gtk_entry_get_text(GTK_ENTRY(ideditor->id_org)));
  if (flags & XEP_IDEDITOR_REPLY_TO_CHANGED)
    a_Identity_set_reply_to(identity_copy,
			    gtk_entry_get_text(GTK_ENTRY(ideditor->id_reply_to)));


  /* XEP_IDEDITOR_DEFAULTS */
  if (flags & XEP_IDEDITOR_DRAFTS_CHANGED)
    a_Identity_set_draft_ref
      (identity_copy, gtk_entry_get_text(GTK_ENTRY(ideditor->def_drafts)));
  if (flags & XEP_IDEDITOR_QUEUE_CHANGED)
    a_Identity_set_queue_ref
      (identity_copy, gtk_entry_get_text(GTK_ENTRY(ideditor->def_queue)));
  if (flags & XEP_IDEDITOR_SENT_CHANGED)
    a_Identity_set_sent_ref
      (identity_copy, gtk_entry_get_text(GTK_ENTRY(ideditor->def_sent)));
  if (flags & XEP_IDEDITOR_REPLY_TO_CHANGED)
    a_Identity_set_templates_ref
      (identity_copy,
       gtk_entry_get_text(GTK_ENTRY(ideditor->def_templates)));

  /* default mta */
  if (flags & XEP_IDEDITOR_MTAS_CHANGED)
    a_Identity_set_mta_ref(identity_copy, 
			   GtkIdEditor_mtas_get_selected(ideditor));
}

#endif /* USE_GTKUI */
