/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef GTK_MESSAGEVIEW_H
#define GTK_MESSAGEVIEW_H

#include <gtk/gtkwidget.h>
#include <gtk/gtktextbuffer.h>
#include <gtk/gtktreestore.h>

#include <libetpan.h>

#include "ui-types.h"
#include "message-types.h"
#include "folder-types.h"
#include "gtk-types.h"
#include "gtk-ui.h"

/* Defined here because ui.h needs it to create the union */
struct XePGTKMessageView_s
{
  /*! The following members must be the same as the ones in
   *  XePUIMessageViewAny defined in ui.c 
   */
  XePUIType type;      /**< UI type */
  XePUID uid;          /**< Unique Identifier for the message view */
  int editable;        /**< Wether editing is allowed */
  XetPan *xetpan;      /**< The global XetPan data structure */
  XePUIMain *uimain;   /**< The parent main window, if any */
  XePFolder *folder;   /**< The parent folder, if any */
  XePMessage *message; /**< Currently displayed message */
  uint32_t muid;       /**< muid of the currently displayed message, needed
			* because message are displayed in chunks,
			* asychronously (incremental reflow)
			*/
  clist *inlined_list; /**< List of XePUtilsInlinedData describing the
			*   inlined data
			*/
  char *body_text;     /**< Body text of composed message */
  char *default_charset; /**< Body's default charset */
  /*! Private to the Gtk+ UI */
  GtkWidget *toplevel; /**< Used when the viewer is a top-level one */
  GtkWidget *widget;
  GtkWidget *text;
  GtkWidget *mime_widget;
  GtkWidget *mime_view;
  GtkWidget *mime_menu;
  GtkWidget *qsearch_box;
  GtkTextBuffer *buffer;
  GtkTreeStore *mime_store;
  int pane_position;

};

#define a_GtkMessageview_show(mview)   gtk_widget_show(mview->widget)
#define a_GtkMessageview_widget(mview) GTK_WIDGET(mview->widget)

XePGTKMessageView *
a_GtkMessageview_create(XetPan *xetpan, 
			GtkWidget *parent, 
			int editable);

XePGTKMessageView *
a_GtkMessageview_create_standalone(XetPan *xetpan, 
				   char *terminal, 
				   int editable);

void
a_GtkMessageview_free(XePGTKMessageView *mview);

void
a_GtkMessageview_link_to_uimain(XePGTKMessageView *mview, XePUIMain *uimain);

void
a_GtkMessageview_message_clear(XePGTKMessageView *mview);

void
a_GtkMessageview_finalize(XePGTKMessageView *mview, XePMessage *message);

void
a_GtkMessageview_mime_view_show(XePGTKMessageView *mview,
				XePMessage *message);

void
a_GtkMessageview_mime_view_block(XePGTKMessageView *mview,
				 XePMessage *message);

void
a_GtkMessageview_mime_view_unblock(XePGTKMessageView *mview,
				   XePMessage *message);

void
a_GtkMessageview_mime_show_selected(XePGTKMessageView *mview,
				    XePMessage *message);

void
a_GtkMessageview_show_headers(XePGTKMessageView *mview, XePMessage *message);

/*! Get the text part of the message to be used for the reply
 *  If text is selected, returns the selected text
 *  If text is marked as "reply", returns a string containing the 
 *  concatenation of the marked parts separated by three empty lines
 */
gchar *
a_GtkMessageview_get_reply_text(XePGTKMessageView *mview);

/*! Get the displayed message
 *  Currenlty, all the text is returned.
 *  TODO: a stucture containing a list of the different text
 *  sections is to be returned. This list will contain the different
 *  text parts and the different inlined files
 */
gchar *
a_GtkMessageview_get_text(XePGTKMessageView *mview);

/*! Set the text of the defautl mime part
 *  \return TRUE on success of FALSE when memory allocation fails
 */
gboolean
a_GtkMessageview_default_part_set_text(XePGTKMessageView *mview, 
				       const char *text, size_t size);

#endif
