/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/poll.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include <gtk/gtkalignment.h>

#include <gdk/gdkkeysyms.h>
#include <gdk/gdkevents.h> /* GDK_BUTTON_PRESS */

#include <clist.h>

#include "intl.h"

#include "x_alloc.h"
#include "defines.h"
#include "logging.h"
#include "remote.h"
#include "pixmaps/xetpan.xpm"
#include "foldersview.h"
#include "ui.h"
#include "gtk-ui.h"
#include "gtk-utils.h"
#include "gtk-log.h"
#include "gtk-about.h"
#include "gtk-foldersview.h"
#include "gtk-messageview.h"
#include "gtk-toolbar.h"
#include "gtk-id-editor.h"
#include "manager.h"
#include "identity.h"
#include "conf.h"
#include "utils.h"
#include "gtkxtreeview.h"

#ifdef USE_PYTHON
#include "gtk-python.h"
#endif

#define XETPAN_TIMES

#ifdef XETPAN_TIMES
#include <sys/times.h>
#endif

#define VIEW_TREE

enum
  {
    SV_NONE,
    SV_STATUS_COL,
    SV_FLAG_COL,
    SV_SUBJECT_COL,
    SV_FROM_COL,
    SV_DATE_COL,
    SV_SIZE_COL,
    SV_TO_COL,
    SV_INDEX_COL
  };

/* TODO: remove it because it is just used to test if gtk was initialized
 * or not
 */
typedef struct XePGTKUI_s XePGTKUI;
struct XePGTKUI_s
{
};

typedef struct XePGtkUIMain_s XePGtkUIMain;
struct XePGtkUIMain_s
{
  XePUIMain ui;
  GtkWidget *copy_folder_selector;
  GtkWidget *move_folder_selector;
  gint pane1_position;
  gint pane2_position;
};

static void
GtkUI_add_stock_icons(void);

static void
GtkUI_icon_factory_add_icon(GtkIconFactory *icon_factory,
			    gchar *icon_name,
			    gchar *stock_id);

/* Main View: menu bar, toolbar, folderview, summaryview, messageview */

static GtkWidget *
GtkUI_mainview_menu_create(XePUIMain *uimain);

static void
GtkUI_mainview_free(XePUIMain *uimain);

static void
GtkUI_messageview_popup_cb(GtkTextView *textview, 
			   GtkMenu *menu, 
			   gpointer data);

static void
GtkUI_mainview_file_new_cb(gpointer data,
			   guint action,
			   GtkWidget *widget);

static void
GtkUI_mainview_file_new_real_cb(void *axetpan,
				int response,
				gchar *display);

static void
GtkUI_mainview_file_close_cb(gpointer data,
			     guint action,
			     GtkWidget *widget);
static void
GtkUI_mainview_edit_quick_search_cb(gpointer data,
				    guint action,
				    GtkWidget *widget);

static void
GtkUI_mainview_edit_flag_set_cb(gpointer data,
				guint flags,
				GtkWidget *widget);

static void
GtkUI_mainview_edit_flag_unset_cb(gpointer data,
				  guint flags,
				  GtkWidget *widget);

static void
GtkUI_mainview_edit_copy_move_cb(gpointer auimain,
				 guint action,
				 GtkWidget *widget);

static void
GtkUI_mainview_edit_identities_cb(gpointer data,
				  guint action,
				  GtkWidget *widget);
#ifdef USE_PYTHON
static void
GtkUI_mainview_tools_python_cb(gpointer data,
			       guint action,
			       GtkWidget *widget);
#endif

static void
GtkUI_mainview_tools_process_cb(gpointer data,
				guint action,
				GtkWidget *widget);

static void
GtkUI_mainview_tools_log_cb(gpointer data,
			    guint action,
			    GtkWidget *widget);

static void
GtkUI_mainview_about_cb(gpointer data,
		     guint action,
		     GtkWidget *widget);

static void
GtkUI_messageview_fetch_source(XePGTKMessageView *mview, 
			       XePManagerMsgCb callback);

static void
GtkUI_messageview_popup_show_source_cb(GtkMenuItem *menu_item,
				       gpointer data);

static void
GtkUI_messageview_popup_save_source_as_cb(GtkMenuItem *menu_item,
					  gpointer data);

static void
GtkUI_messageview_show_source_cb(void *amview, void *amessage);

static void
GtkUI_messageview_save_source_cb(void *amview, void *amessage);

static gboolean
GtkUI_mainview_delete_cb(GtkWidget *widget,
			 GdkEvent *event,
			 gpointer data);

/* Folder View */

static gboolean
GtkUI_foldersview_button_release_cb(XePGtkFoldersView *fview,
				    GdkEventButton *event,
				    gpointer auimain);
/* Summary View */

static XePGTKSummaryView *
GtkUI_summaryview_new(XePUIMain *uimain);

static void
GtkUI_summaryview_free(XePGTKSummaryView *sview);

static void
GtkUI_summaryview_quick_search_show(XePGTKSummaryView *sview);

static void
GtkUI_summaryview_quick_search_changed_cb(GtkEditable *editable,
					  gpointer asview);
static gboolean
GtkUI_summaryview_quick_search_key_press_cb(GtkWidget *widget,
					    GdkEventKey *event,
					    gpointer asview);
static gboolean
GtkUI_summaryview_quick_search_func(void *iterator, gpointer asview);

static gboolean
GtkUI_summaryview_match_new_func(void *iterator, 
				  gpointer asview);

static gboolean
GtkUI_summaryview_msg_data_func(mailmessage *msg, 
				guint col_id,
				void **content,
				gchar **attribute,
				gpointer data);


static gboolean
GtkUI_summaryview_tree_data_func(void *iterator, 
				guint col_id,
				void **content,
				gchar **attribute,
				gpointer data);

static void
GtkUI_summaryview_tree_params_func(void *iterator, 
				  gboolean *has_children,
				  guint *depth,
				  gpointer data);

static gboolean
GtkUI_summaryview_tree_iterator_func(void **iterator, 
				     gboolean next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data);
#if 0
static gboolean
GtkUI_summaryview_list_data_func(void *iterator, 
				 guint col_id,
				 void **content,
				 gchar **attribute,
				 gpointer data);

static void
GtkUI_summaryview_list_params_func(void *iterator, 
				  gboolean *has_children,
				  guint *depth,
				  gpointer data);

static gboolean
GtkUI_summaryview_list_iterator_func(void **iterator, 
				     gboolean next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data);
#endif

static mailmessage *
GtkUI_summaryview_get_selected_msg(XePGTKSummaryView *sview);

static void
GtkUI_summaryview_row_activated_cb(GtkXTreeView *xtv, 
				   void *iterator, 
				   guint col_id, 
				   gpointer asview);

static gboolean
GtkUI_summaryview_button_release_cb(GtkWidget *widget,
				    GdkEventButton *even,
				    gpointer data);
static void
GtkUI_summaryview_set_view(XePGTKSummaryView *sview,
			   XePFolderRef *folder_ref,
			   gboolean update_folder);

static void
GtkUI_summaryview_refresh(XePGTKSummaryView *sview);

static void
GtkUI_summaryview_flag_selection(XePGTKSummaryView *sview, 
				 XePMessageFlagsOp flag_op,
				 int flags);

static void
GtkUI_summaryview_copy_done_cb(carray *folders, 
			       carray *messages, 
			       uint32_t failures,
			       void *asview);

static void
GtkUI_summaryview_copy_move_selection(XePGTKSummaryView *sview,
				      gboolean move);

/* libEtPan! extensions */
static struct mail_envelopes_tree *
mail_envelopes_tree_next(struct mail_envelopes_tree *self,
			 struct mail_envelopes_tree *child,
			 int ignore_children);
static struct mail_envelopes_tree *
mail_envelopes_tree_prev(struct mail_envelopes_tree *parent,
			 struct mail_envelopes_tree *self);


#if 0
static uint32_t
mail_envelopes_list_next(struct mail_envelopes_list *ep_list, uint32_t index);

static uint32_t
mail_envelopes_list_prev(struct mail_envelopes_list *ep_list, uint32_t index);
#endif

/* IO Watch Callback */

static gboolean
GtkUI_io_watch_cb(GIOChannel *channel, GIOCondition gcondition, gpointer data);

/******************************************************************************/
/*                                                                            */
/* Basic GUI functions                                                        */
/*                                                                            */
/******************************************************************************/

/* Init display and abort if fail */
void
a_GtkUI_init(XetPan *xetpan, int *argc, char ***argv)
{
  A_LOGGING_TRACE;
  if (a_GtkUI_init_check(xetpan, argc, argv))
    {
      A_LOGGING_ERROR("Unable to initilialize GUI. Aborting");
      exit(1);
    }
}

/* Init display and return -1 if fail */
int
a_GtkUI_init_check(XetPan *xetpan, int *argc, char ***argv)
{
  int ret;

  A_LOGGING_TRACE;

  ret = 0;

  if (gtk_init_check(argc, argv) == FALSE)
    {
      const char *display_name_arg = gdk_get_display_arg_name ();
      A_LOGGING_ERROR("GTK+ User Interface could not be initialized:");
      A_LOGGING_ERROR("cannot open display %s", 
		      display_name_arg ? display_name_arg : " ");
      ret = -1;
    }
  else
    {
      if (xetpan->ui->gtk)
	{
	  A_LOGGING_INFO("GTK+ User Interface already initialized.");
	}
      else
	{
	  xetpan->ui->gtk = x_alloc(1, XePGTKUI);
	  if (xetpan->ui->gtk == NULL)
	    {
	      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
	      ret = -1;
	    }
          else
            {
              gtk_settings_set_long_property(gtk_settings_get_default(),
                                             "gtk-can-change-accels",
                                             1, 
                                             "XetPan");
	      GtkUI_add_stock_icons();
	      A_LOGGING_INFO("Initialized GTK+ User Interface");
            }
	}
    }
  return ret;
}

/* Main loop */
void
a_GtkUI_main(XetPan *xetpan)
{
  int n_watches, i;
  XePUIIOWatcher *watcher;
  XePGTKUI *gtkui = (XePGTKUI *) xetpan->ui->gtk;

  A_LOGGING_TRACE;
  
  gtk_main();
  
  n_watches = carray_count(xetpan->ui->watched_fd);
  for (i = 0; i < n_watches; i++)
    {
      watcher = (XePUIIOWatcher *) carray_get(xetpan->ui->watched_fd, i);
      a_GtkUI_delete_io_watch(xetpan->ui, watcher);
    }

  x_free(gtkui);
  xetpan->ui->gtk = NULL;
  
  A_LOGGING_DEBUG("xetpan->ui->gtk = %p", xetpan->ui->gtk);

  A_LOGGING_DEBUG("end");
}

void
a_GtkUI_main_quit(XetPan *xetpan)
{
  A_LOGGING_TRACE;
  gtk_main_quit();
}

int
a_GtkUI_events_pending(XetPan *xetpan)
{
  return gtk_events_pending();
}

void
a_GtkUI_main_iteration(XetPan *xetpan)
{
  gtk_main_iteration();
}

static void
GtkUI_add_stock_icons(void)
{
  GtkIconFactory *icon_factory;

  icon_factory = gtk_icon_factory_new();
  

  GtkUI_icon_factory_add_icon(icon_factory, "xetpan.png",
			      XEP_STOCK_ICON);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-status.png",
			      XEP_STOCK_STATUS);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-flag.png",
			      XEP_STOCK_FLAG);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-reply.png",
			      XEP_STOCK_REPLY);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-forward.png",
			      XEP_STOCK_FORWARD);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-replied.png",
			      XEP_STOCK_REPLIED);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-forwarded.png",
			      XEP_STOCK_FORWARDED);
  GtkUI_icon_factory_add_icon(icon_factory, "xetpan-replied-forwarded.png",
			      XEP_STOCK_REPLIED_FORWARDED);

  gtk_icon_factory_add_default(icon_factory);
}

static void
GtkUI_icon_factory_add_icon(GtkIconFactory *icon_factory,
			    gchar *icon_name,
			    gchar *stock_id)
{
  GtkIconSet *icon_set;
  GtkIconSource *icon_source;
  gchar *filename;

  icon_set = gtk_icon_set_new();

  icon_source = gtk_icon_source_new();

  filename = g_strdup_printf("%s/pixmaps/%s", PKGDATADIR, icon_name);

  gtk_icon_source_set_filename(icon_source, filename);

  g_free(filename);

  gtk_icon_set_add_source(icon_set, icon_source);

  gtk_icon_factory_add(icon_factory, stock_id, icon_set);
  gtk_icon_source_free(icon_source);

}


/******************************************************************************/
/*                                                                            */
/* Windows                                                                    */
/*                                                                            */
/******************************************************************************/

void
a_GtkUI_first_launch(XetPan *xetpan)
{
  A_LOGGING_DEBUG("not implemented yet");
}

void
a_GtkUI_compose_pending_requests(XetPan *xetpan)
{
  A_LOGGING_DEBUG("not implemented yet");
}

/** Create a new Gtk+ main view 
 *  It should not be called directly: it must be called by
 *  a_UI_mainview_create() so the mainview is registered
 */
XePUIMain *
a_GtkUI_mainview_create(XetPan *xetpan, char *terminal)
{
  XePGtkUIMain *gtkuimain;
  XePUIMain *uimain;
  XePGtkFoldersView *fview;
  XePGTKSummaryView *sview;
  XePGTKMessageView *mview;
  GtkWidget *toplevel;
  GtkWidget *topvbox;
  GtkWidget *menubar;
#if WITH_TOOLBAR_HANDLE
  GtkWidget *toolbar_handle;
#endif
  GtkWidget *toolbar_hbox;
  GtkWidget *main_frame;
  GtkWidget *fview_frame;
  GtkWidget *message_frame;
  GtkWidget *vertical_pane;
  GtkWidget *horizontal_pane;

  gtkuimain = x_alloc(1, XePGtkUIMain);

  if (gtkuimain == NULL)
    return NULL;

  uimain = (XePUIMain *) gtkuimain;

  /* TODO: clean up */
  gtkuimain->copy_folder_selector = NULL;
  gtkuimain->move_folder_selector = NULL;

  a_UI_mainview_init((XePUIMain *) uimain);

  uimain->type = XETPAN_UI_GTK;
  uimain->xetpan = xetpan;

  toplevel = a_GtkUtils_toplevel_new(terminal,
				     _("XetPan"), 
				     xetpan_xpm,
				     620, 460);
  uimain->toplevel = toplevel;

  topvbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(topvbox);
  gtk_container_add(GTK_CONTAINER(toplevel), topvbox);

  /* Menu bar */

  menubar = GtkUI_mainview_menu_create(uimain);
  gtk_widget_show(menubar);
  gtk_box_pack_start(GTK_BOX(topvbox), menubar, FALSE, FALSE, 0);

  uimain->menubar = menubar;

  /* Toolbar */

#ifdef  WITH_TOOLBAR_HANDLE
  toolbar_handle = gtk_handle_box_new();
  gtk_widget_show(toolbar_handle);
  gtk_handle_box_set_shadow_type(GTK_HANDLE_BOX(toolbar_handle), 
				 GTK_SHADOW_OUT);
  gtk_box_pack_start(GTK_BOX(topvbox), toolbar_handle, FALSE, FALSE, 0);
#endif

  toolbar_hbox = a_GtkToolbar_mainview_new(uimain);
  gtk_widget_show(toolbar_hbox);

#ifdef WITH_TOOLBAR_HANDLE
  gtk_container_add(GTK_CONTAINER(toolbar_handle), toolbar_hbox);
#else
  gtk_box_pack_start(GTK_BOX(topvbox), toolbar_hbox, FALSE, FALSE, 0);
#endif

  /* Main Frame */

  main_frame = gtk_frame_new(NULL);
  gtk_widget_show(main_frame);
  gtk_frame_set_shadow_type(GTK_FRAME(main_frame), GTK_SHADOW_OUT);
  gtk_box_pack_start(GTK_BOX(topvbox), main_frame, TRUE, TRUE, 0);
  
  /* Horizontal Pane */

  horizontal_pane = gtk_hpaned_new();
  gtk_widget_show(horizontal_pane);

  /* Vertical Pane */

  vertical_pane = gtk_vpaned_new();
  gtk_widget_show(vertical_pane);

#ifndef NEW_PANES
  gtk_container_add(GTK_CONTAINER(main_frame), horizontal_pane);
  gtk_paned_add2(GTK_PANED(horizontal_pane), vertical_pane);
#else
  gtk_container_add(GTK_CONTAINER(main_frame), vertical_pane);
  gtk_paned_add2(GTK_PANED(vertical_pane), horizontal_pane);
#endif

  /* Folders view */
 
  /* FIXME: check if fview is NULL */
  fview_frame = gtk_alignment_new(0.0, 0.0, 1.0, 1.0);

  gtk_widget_show(fview_frame);

  gtk_paned_add1(GTK_PANED(horizontal_pane), fview_frame);
  
  fview = a_GtkFoldersview_new(GTK_SELECTION_SINGLE);

  uimain->foldersview = fview;

  gtk_widget_show(XEP_GTK_FOLDERSVIEW_WIDGET(fview));

  gtk_container_add(GTK_CONTAINER(fview_frame),
		    XEP_GTK_FOLDERSVIEW_WIDGET(fview));

  /* Summary view */

  /* FIXME: check if sview is NULL */
  sview = GtkUI_summaryview_new(uimain);
  uimain->summaryview = sview;

  gtk_paned_pack1(GTK_PANED(vertical_pane),
		  a_GtkUI_summaryview_widget(sview), TRUE, TRUE);

  a_GtkUI_summaryview_show(sview);

  /* Message view */

  message_frame = gtk_alignment_new(0.0, 0.0, 1.0, 1.0);
  gtk_widget_show(message_frame);

#ifndef NEW_PANES
  gtk_paned_pack2(GTK_PANED(vertical_pane), message_frame, TRUE, TRUE);
#else
  gtk_paned_add2(GTK_PANED(horizontal_pane), message_frame);
#endif

  mview = a_GtkMessageview_create(uimain->xetpan, message_frame, FALSE);
  a_GtkMessageview_link_to_uimain(mview, uimain);

  uimain->messageview = mview;

  /* Python Console */

  uimain->python = NULL;

  /* Log View */

  uimain->logview = a_GtkLog_new(uimain->xetpan, toplevel);

  /* TODO: pane position should be restored from preferences */
  {
    GtkPaned *paned = (GTK_PANED(vertical_pane));
    gtk_paned_set_position(paned, 150);
  }

  /* Signals */

  a_GtkFoldersview_set_button_release_callback
    (fview, GtkUI_foldersview_button_release_cb, uimain);

  g_signal_connect(G_OBJECT(toplevel),
		   "delete-event",
		   G_CALLBACK(GtkUI_mainview_delete_cb),
		   uimain);

  g_signal_connect(G_OBJECT(mview->text), "populate-popup",
		   G_CALLBACK(GtkUI_messageview_popup_cb),
		   mview);

  a_GtkUtils_paned_shrink_control(horizontal_pane, 
				  &(gtkuimain->pane1_position));

  a_GtkUtils_paned_shrink_control(vertical_pane, &(gtkuimain->pane2_position));

  gtk_widget_show(toplevel); 

  a_GtkFoldersview_set_list(fview,
			    xetpan->config->configuration->foldersviewslist);

  a_GtkMessageview_show(mview);

  return uimain;
}

static gboolean
GtkUI_mainview_delete_cb(GtkWidget *widget,
			 GdkEvent *event,
			 gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;

  A_LOGGING_TRACE;
  if (a_UI_mainview_close(uimain))
    {
      a_XetPan_quit(xetpan);
    }
  return FALSE;
}

void
a_GtkUI_mainview_close(XePUIMain *uimain)
{
  A_LOGGING_TRACE;
  GtkUI_mainview_free(uimain);
}

static void
GtkUI_mainview_free(XePUIMain *uimain)
{
  A_LOGGING_TRACE;
  if (uimain->type != XETPAN_UI_GTK)
    {
      A_LOGGING_SYSTEM_ERROR("Attempt to close a non Gtk+ UI with "
			     "a Gtk+ function");
      return;
    }

  GtkUI_summaryview_free((XePGTKSummaryView *) uimain->summaryview);
  a_UI_messageview_close((XePUIMessageView *) uimain->messageview);
  a_GtkToolbar_mainview_free(uimain);

  /* All the rest is freed when freeing the toplevel widget */
  gtk_widget_destroy((GtkWidget *) uimain->toplevel);
  x_free(uimain);
}

static GtkWidget *
GtkUI_mainview_menu_create(XePUIMain *uimain)
{
  GtkAccelGroup *accel_group;
  GtkItemFactory *item_factory;
  GtkItemFactoryEntry entries[] = {
    { N_("/_File"), NULL, NULL, 0, "<Branch>" },
    { N_("/File/_New Window"), "<Control>N", GtkUI_mainview_file_new_cb, 0,
      NULL},
    { N_("/File/New Window in _Other Display..."), NULL, 
      GtkUI_mainview_file_new_cb, 1, NULL},
    { N_("/File/_Close"), "<Control>W", GtkUI_mainview_file_close_cb, 0,
      "<StockItem>", GTK_STOCK_CLOSE},
    { N_("/File/_Quit"), "<Control>Q", GtkUI_mainview_file_close_cb, 1,
      "<StockItem>", GTK_STOCK_QUIT},  
    { N_("/_Edit"), NULL, NULL, 0, "<Branch>" },
    { N_("/Edit/Quick Search in Folder..."), NULL, 
      GtkUI_mainview_edit_quick_search_cb, 0, "<StockItem>", GTK_STOCK_FIND},
    { N_("/Edit/Quick Search in Message..."), NULL, 
      GtkUI_mainview_edit_quick_search_cb, 1, NULL},
    { "/Edit/sep1", NULL, NULL, 0, "<Separator>"},
    { N_("/Edit/_Mark Message"), NULL, NULL, 0, "<Branch>"},
    { N_("/Edit/Mark Message/Set Deleted"), "<Ctrl>d",
      GtkUI_mainview_edit_flag_set_cb, MAIL_FLAG_DELETED, 
      "<StockItem>", XEP_STOCK_DELETED},      
    { N_("/Edit/Mark Message/Set Flagged"), NULL,
      GtkUI_mainview_edit_flag_set_cb, MAIL_FLAG_FLAGGED, 
      "<StockItem>", XEP_STOCK_FLAG},      
    { N_("/Edit/Mark Message/Set Unseen"), "<Ctrl>u",
      GtkUI_mainview_edit_flag_unset_cb, MAIL_FLAG_SEEN,
      "<StockItem>", XEP_STOCK_UNSEEN},
    { N_("/Edit/Mark Message/Set Answered"), NULL,
      GtkUI_mainview_edit_flag_set_cb, MAIL_FLAG_ANSWERED, 
      "<StockItem>", XEP_STOCK_REPLIED},
    { N_("/Edit/Mark Message/Set Forwarded"), NULL,
      GtkUI_mainview_edit_flag_set_cb, MAIL_FLAG_FORWARDED, 
      "<StockItem>", XEP_STOCK_FORWARDED},
    { "/Edit/Mark Message/sep1", NULL, NULL, 0, "<Separator>"},
    { N_("/Edit/Mark Message/Unset Deleted"), "<Ctrl><Shift>d",
      GtkUI_mainview_edit_flag_unset_cb, MAIL_FLAG_DELETED, NULL},
    { N_("/Edit/Mark Message/Unset Flagged"), NULL,
      GtkUI_mainview_edit_flag_unset_cb, MAIL_FLAG_FLAGGED, NULL},
    { N_("/Edit/Mark Message/Unset Unseen"), "<Ctrl><Shift>u",
      GtkUI_mainview_edit_flag_set_cb, MAIL_FLAG_SEEN, NULL},
    { N_("/Edit/Mark Message/Unset Answered"), NULL,
      GtkUI_mainview_edit_flag_unset_cb, MAIL_FLAG_ANSWERED, NULL},
    { N_("/Edit/Mark Message/Unset Forwarded"), NULL,
      GtkUI_mainview_edit_flag_unset_cb, MAIL_FLAG_FORWARDED, NULL},
    { "/Edit/sep2", NULL, NULL, 0, "<Separator>"},
    { N_("/Edit/_Copy to Folder..."), NULL, 
      GtkUI_mainview_edit_copy_move_cb, 0, NULL},
    { N_("/Edit/_Move to Folder..."), NULL,
      GtkUI_mainview_edit_copy_move_cb, 1, NULL},
    { "/Edit/sep3", NULL, NULL, 0, "<Separator>"},
    { N_("/Edit/E_xpunge Folder"), "<Ctrl><Shift>x",
      a_GtkUI_summaryview_expunge, 0, NULL},
    { "/Edit/sep3", NULL, NULL, 0, "<Separator>"},
    { N_("/Edit/_Identities..."), NULL, GtkUI_mainview_edit_identities_cb, 0,
      NULL},
    { N_("/Edit/Prefere_nces..."), NULL, NULL, 0, "<StockItem>", 
      GTK_STOCK_PREFERENCES},
    { N_("/_View"), NULL, NULL, 0, "<Branch>" },
    { N_("/_Windows"), NULL, NULL, 0, "<Branch>" },
    { N_("/_Tools"), NULL, NULL, 0, "<Branch>" },
#if USE_PYTHON
    { N_("/Tools/_Python Console"), NULL, GtkUI_mainview_tools_python_cb, 0},
#endif
    { N_("/Tools/Process"), NULL, GtkUI_mainview_tools_process_cb, 0},
    { N_("/Tools/_Log Viewer"), NULL, GtkUI_mainview_tools_log_cb, 0, NULL},
    { N_("/_Help"), NULL, NULL, 0, "<Branch>" },
    { N_("/Help/_Contents"), "F1", NULL, 0, "<StockItem>", GTK_STOCK_HELP},
    { N_("/Help/_About"), NULL, GtkUI_mainview_about_cb, 0, NULL}};


  accel_group = gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<Main>", 
				      accel_group);
  gtk_item_factory_set_translate_func(item_factory, gettext, NULL, NULL);
  gtk_item_factory_create_items(item_factory, 
				sizeof(entries)/sizeof(entries[0]),
				entries, uimain);
  gtk_window_add_accel_group(GTK_WINDOW(uimain->toplevel), accel_group);
  return gtk_item_factory_get_widget(item_factory, "<Main>");
}

static void
GtkUI_messageview_popup_cb(GtkTextView *textview, GtkMenu *menu, gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkWidget *menu_item;

  if (mview->message && mview->message->muid == mview->muid)
    {
      menu_item = gtk_menu_item_new_with_label(_("Show message source"));
      g_signal_connect(G_OBJECT(menu_item), "activate",
		       G_CALLBACK(GtkUI_messageview_popup_show_source_cb),
		       mview);
      gtk_menu_append(menu, menu_item);
      gtk_widget_show(menu_item);

      menu_item = gtk_menu_item_new_with_label(_("Save message source as..."));
      g_signal_connect(G_OBJECT(menu_item), "activate",
		       G_CALLBACK(GtkUI_messageview_popup_save_source_as_cb),
		       mview);
      gtk_menu_append(menu, menu_item);
      gtk_widget_show(menu_item);

    }
}

static void
GtkUI_mainview_file_new_cb(gpointer data,
			   guint action,
			   GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;
  gchar *display;

  display =  gdk_screen_make_display_name(gtk_window_get_screen
					      (GTK_WINDOW(uimain->toplevel)));
  if (action)
    {
      a_GtkUtils_dialog_get_string(_("Open New Window in Other Display"),
				   uimain->toplevel,
				   _("Display Name:"),
				   NULL,
				   display,
				   XEP_GTKUTILS_OK | XEP_GTKUTILS_CANCEL,
				   FALSE,
				   GtkUI_mainview_file_new_real_cb,
				   xetpan);
    }
  else
    {
      a_UI_mainview_create(xetpan, display, XETPAN_UI_GTK);
    }

  g_free(display);     
}

static void
GtkUI_mainview_file_new_real_cb(void *axetpan,
				int response,
				gchar *display)
{
  XetPan *xetpan = (XetPan *) axetpan;

  if (response == GTK_RESPONSE_OK)
    {
      a_UI_mainview_create(xetpan, display, XETPAN_UI_GTK);
    }
}

static void
GtkUI_mainview_file_close_cb(gpointer data,
			     guint action,
			     GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XetPan *xetpan = (XetPan *) uimain->xetpan;
  if (action || a_UI_mainview_close(uimain))
    a_XetPan_quit(xetpan);
}

static void
GtkUI_mainview_edit_quick_search_cb(gpointer data,
				    guint action,
				    GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  switch (action)
    {
    case 0:
      GtkUI_summaryview_quick_search_show(sview);
      break;
    default:
      break;
    }
}

static void
GtkUI_mainview_edit_flag_set_cb(gpointer data,
				guint flags,
				GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  GtkUI_summaryview_flag_selection(sview, XEP_FLAGS_SET, flags);
}

static void
GtkUI_mainview_edit_flag_unset_cb(gpointer data,
				  guint flags,
				  GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  GtkUI_summaryview_flag_selection(sview, XEP_FLAGS_UNSET, flags);
}

static void
GtkUI_mainview_edit_copy_move_cb(gpointer auimain,
				 guint action,
				 GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) auimain;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  GtkUI_summaryview_copy_move_selection(sview, action == 1);
}

static void
GtkUI_mainview_edit_identities_cb(gpointer data,
				  guint action,
				  GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  a_GtkIdEditor_show(uimain);
}

#ifdef USE_PYTHON
static void
GtkUI_mainview_tools_python_cb(gpointer data,
			       guint action,
			       GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKPython *xeppy = (XePGTKPython *) uimain->python;

  if (xeppy == NULL)
    uimain->python = a_GtkPython_console_new(uimain->xetpan, 
					     GTK_WIDGET(uimain->toplevel));
  else
    a_GtkPython_console_show(xeppy);
}
#endif

static void
GtkUI_mainview_tools_process_cb(gpointer data,
				guint action,
				GtkWidget *widget)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  GtkWidget *toplevel = (GtkWidget *) sview->uimain->toplevel;
  XePFolder *folder = sview->folder_ref->folder;
  void **selection;
  uint32_t n;
  clist *processing_rules;
  
  /* TODO: Processing menu item must be inactivated when no folder is
   * selected */
  processing_rules = a_Folder_get_processing_rules(folder);
  if (processing_rules == NULL)
    {
      a_GtkUtils_dialog_alert(_("Folder Processing"), toplevel,
			      XETPAN_ERROR_MESSAGE,
			      _("No Processing Rules"),
			      _("The folder does not have any processing rules.\n"
				"Processing rules can be defined in "
				"<bold>Edit/Processing Rules...</bold>"),
			      XEP_GTKUTILS_OK,
			      TRUE,
			      TRUE,
			      NULL, NULL);
      return;
    }

  n = gtk_xtree_view_get_selection(GTK_XTREE_VIEW(sview->tree), &selection);

  if (n)
    { 
      carray *messages;
      
      gtk_xtree_view_freeze(GTK_XTREE_VIEW(sview->tree));
      
      messages = a_Message_array_new(folder, selection, n);
      
      free(selection);

      a_UI_folder_process(folder, messages, processing_rules);
    }
}

static void
GtkUI_mainview_tools_log_cb(gpointer data,
			    guint action,
			    GtkWidget *widget)
{
  a_GtkLog_show((XePGTKLog *)((XePUIMain *) data)->logview);
}

static void
GtkUI_mainview_about_cb(gpointer data,
			guint action,
			GtkWidget *widget)
{
  a_GtkUI_about((XePUIMain *) data);
}

static void
GtkUI_messageview_fetch_source(XePGTKMessageView *mview, 
			       XePManagerMsgCb callback)
{
  XePManagerMsg *msg;
  
  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_FETCH);
  a_Manager_msg_set_source_target(msg, NULL, &mview->message->folder->thread);
  a_Manager_msg_set_args(msg, mview->message, mview->message);
  a_Manager_msg_add_callback(msg, callback, mview);
  a_Manager_msg_submit(msg);
}

static void
GtkUI_messageview_popup_show_source_cb(GtkMenuItem *menu_item, gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkUI_messageview_fetch_source(mview, (XePManagerMsgCb) GtkUI_messageview_show_source_cb);
}

static void
GtkUI_messageview_show_source_cb(void *amview, void *amessage)
{
  char *display;
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  XePMessage *message = (XePMessage *) amessage;

  display = a_GtkUtils_make_display_name(mview->toplevel);

  a_GtkUtils_file_viewer_new(display, _("Message source"),
			     NULL, 0, 0, NULL, message->data, message->len,
			     message->source_charset);
  x_free(display);
  x_free(message->data);
  message->data = NULL;
  message->len  = 0; 
}

static void
GtkUI_messageview_popup_save_source_as_cb(GtkMenuItem *menu_item,
					  gpointer data)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) data;
  GtkUI_messageview_fetch_source(mview, (XePManagerMsgCb) GtkUI_messageview_save_source_cb);
}

static void
GtkUI_messageview_save_source_cb(void *amview, void *amessage)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  XePMessage *message = (XePMessage *) amessage;

  a_GtkUtils_file_selection(_("Save message source as..."),
			    mview->toplevel,
			    "email.txt",
			    (XePGtkUtilsSelectionCb) a_Message_data_save_as,
			    message);
			    
}

void
a_GtkUI_mainview_reply(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePGTKMessageView *mview = (XePGTKMessageView *) uimain->messageview;
  XePMessage *message;
  gchar *display;
  mailmessage *msg;
  /* Create XePMessage(s) from the selected messages,
   * retrieve their default part, or the one that is
   * currently viewed, then forge a reply
   * This should be done in the ui.c
   */
  msg = GtkUI_summaryview_get_selected_msg(sview);
  if (msg)
    {
      gchar *reply_text;
      const char *reply_charset = NULL;

      message = a_Message_new_for_fetching(sview->folder_ref->folder, msg);
      display = a_GtkUtils_make_display_name(GTK_WIDGET(uimain->toplevel));

      /* TODO: Reply text should be retrieved from the UI 
       * to allow for selective text reply.
       * If the UI does not provide the text to reply to,
       * then use the default MIME part of the message (this
       * is done in ui.c
       */
      /* TODO: Make sure the message viewed is the one to be 
       * replied to.
       */
      reply_text = a_GtkMessageview_get_reply_text(mview);
      if (reply_text )
	{
	  reply_charset = GTK_DEST_CHARSET;
	}

      a_UI_composewindow_reply(uimain->xetpan,
			       display, 
			       uimain->type, 
			       message,
			       reply_text,
			       reply_charset);
      g_free(reply_text);
      g_free(display);
    }
}

void
a_GtkUI_mainview_forward(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePMessage *message;
  gchar *display;
  mailmessage *msg;
  /* Create XePMessage(s) from the selected messages,
   * retrieve their default part, or the one that is
   * currently viewed, then forge a reply
   * This should be done in the ui.c
   */
  msg = GtkUI_summaryview_get_selected_msg(sview);
  if (msg)
    {
      message = a_Message_new_for_fetching(sview->folder_ref->folder, msg);
      display = a_GtkUtils_make_display_name(GTK_WIDGET(uimain->toplevel));
      
      a_UI_composewindow_forward(uimain->xetpan, 
				 display, 
				 uimain->type,
				 message);
      g_free(display);
    }
}

void
a_GtkUI_mainview_redirect(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePMessage *message;
  gchar *display;
  mailmessage *msg;
  /* Create XePMessage(s) from the selected messages,
   * retrieve their default part, or the one that is
   * currently viewed, then forge a reply
   * This should be done in the ui.c
   */
  msg = GtkUI_summaryview_get_selected_msg(sview);
  if (msg)
    {
      message = a_Message_new_for_fetching(sview->folder_ref->folder, msg);
      display = a_GtkUtils_make_display_name(GTK_WIDGET(uimain->toplevel));
      
      a_UI_composewindow_redirect(uimain->xetpan, 
				  display, 
				  uimain->type,
				  message);
      
      g_free(display);
    }
}

void
a_GtkUI_about(XePUIMain *uimain)
{
  a_GtkAbout_show(uimain);
}

void
a_GtkUI_notify_error(XePUIMain *uimain)
{
  a_GtkToolbar_status_notify_error(uimain);
}

void
a_GtkUI_notify_error_reset(XePUIMain *uimain)
{
  a_GtkToolbar_status_notify_error_reset(uimain);
}

void
a_GtkUI_message_updated(XePUIMain *uimain, XePMessage *message)
{
  if (message && message->folder && uimain && uimain->summaryview)
    {
      XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
      gtk_xtree_view_update(GTK_XTREE_VIEW(sview->tree));
    }
}

void
a_GtkUI_foldersview_list_updated(XePUIFoldersView *fview)
{
  a_GtkFoldersview_list_updated((XePGtkFoldersView *) fview);
}

void
a_GtkUI_foldersview_list_refresh(XePUIFoldersView *fview)
{
  a_GtkFoldersview_list_refresh((XePGtkFoldersView *) fview);
}

static gboolean
GtkUI_foldersview_button_release_cb(XePGtkFoldersView *fview,
				    GdkEventButton *event,
				    gpointer auimain)
{
  XePUIMain *uimain = (XePUIMain *) auimain;
  XePGTKSummaryView *summaryview = (XePGTKSummaryView *) uimain->summaryview;

  if (event->button == 1 || event->button == 2)
    {
      XePFolderRef *folder_ref;
      folder_ref = a_GtkFoldersview_get_selected(fview);
      if (folder_ref)
	{
	  GtkUI_summaryview_set_view(summaryview, folder_ref, 
				     event->button == 2);
	}
      return TRUE;
    }
  return TRUE;
}

/* Summary View */
static XePGTKSummaryView *
GtkUI_summaryview_new(XePUIMain *uimain)
{
  XePGTKSummaryView *sview;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *tree;
  GtkWidget *image;
  GtkWidget *eventbox;

  sview = x_alloc(1, XePGTKSummaryView);
  if (!sview)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }

  sview->uimain = uimain;
  sview->folder_ref = a_Folder_ref_new();

  sview->widget = vbox = gtk_vbox_new(FALSE, 0);
 
  /* The real summary view widge */

  sview->tree = tree = gtk_xtree_view_new();
  gtk_xtree_view_set_match_func(GTK_XTREE_VIEW(tree), 
				GtkUI_summaryview_match_new_func,
				sview);
  gtk_box_pack_start(GTK_BOX(vbox), tree, TRUE, TRUE, 0);

  /* New */

  image = gtk_image_new_from_stock(XEP_STOCK_NEW, GTK_ICON_SIZE_MENU);
  gtk_widget_show(image);
  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_STATUS_COL,
			       GTK_XCOLUMN_TYPE_STOCKID, 
			       TRUE, GTK_SORT_ASCENDING, image, NULL);
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_STATUS_COL, 3*8);
			       
  /* Flagged */
  image = gtk_image_new_from_stock(XEP_STOCK_FLAG, GTK_ICON_SIZE_MENU);
  gtk_widget_show(image);
  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_FLAG_COL,
			       GTK_XCOLUMN_TYPE_STOCKID, 
			       TRUE, GTK_SORT_ASCENDING, image, NULL);
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_FLAG_COL, 3*8);

  /* Subject column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_SUBJECT_COL,
			       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("Subject"));
  
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_SUBJECT_COL, 14*8);
  gtk_xtree_view_set_expander_column(GTK_XTREE_VIEW(tree), SV_SUBJECT_COL);
  gtk_xtree_view_set_expansible_column(GTK_XTREE_VIEW(tree), SV_SUBJECT_COL);


  /* From column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_FROM_COL,
			       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("From"));
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_FROM_COL, 14*8);
  
  /* Date column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_DATE_COL,
		       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("Date"));
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_DATE_COL, 14*8);

  /* Size column */

  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(tree), 0, 
			       SV_SIZE_COL,
			       GTK_XCOLUMN_TYPE_TEXT, 
			       TRUE, GTK_SORT_ASCENDING, NULL, _("Size"));
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(tree), SV_SIZE_COL, 5*8);

  /* Data function */
#ifdef VIEW_TREE
  gtk_xtree_view_set_row_data_func(GTK_XTREE_VIEW(tree), 
				   GtkUI_summaryview_tree_data_func,
				   sview);

  gtk_xtree_view_set_row_params_func(GTK_XTREE_VIEW(tree), 
				     GtkUI_summaryview_tree_params_func,
				     sview);
  gtk_xtree_view_set_data_iterator_func(GTK_XTREE_VIEW(tree), 
					GtkUI_summaryview_tree_iterator_func,
					sview);
#else
  gtk_xtree_view_set_row_data_func(GTK_XTREE_VIEW(tree), 
				   GtkUI_summaryview_list_data_func,
				   sview);

  gtk_xtree_view_set_row_params_func(GTK_XTREE_VIEW(tree), 
				     GtkUI_summaryview_list_params_func,
				     sview);
  gtk_xtree_view_set_data_iterator_func(GTK_XTREE_VIEW(tree), 
					GtkUI_summaryview_list_iterator_func,
					sview);
#endif

  /* The filtering and search box */

  sview->valid_preg = FALSE;

  sview->search_box = hbox = gtk_hbox_new(FALSE, 0);

  sview->search_icon = image = gtk_image_new_from_stock(GTK_STOCK_FIND, 
							GTK_ICON_SIZE_MENU);
  gtk_widget_show(image);

  eventbox = gtk_event_box_new();
  gtk_container_add(GTK_CONTAINER(eventbox), image);
  gtk_widget_show(eventbox);

  a_GtkUtils_set_tip(eventbox, _("s REGEXP: match subject\n"
			      "f REGEXP: match from\n"
			      "d [<>]DATE: match before/after date\n"
			      "n: match new\n"
			      "u: match unread\n"
			      "f FLAG: match flag"));
  gtk_box_pack_start(GTK_BOX(hbox), eventbox, FALSE, TRUE, 12);
  sview->search_entry = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox), sview->search_entry, TRUE, TRUE, 0);
  gtk_widget_show(sview->search_entry);
  gtk_box_pack_start(GTK_BOX(sview->widget), sview->search_box, 
		     FALSE, FALSE, 0);

  g_signal_connect(G_OBJECT(tree), "row-activated",
		   G_CALLBACK(GtkUI_summaryview_row_activated_cb),
		   sview);

  g_signal_connect_after(G_OBJECT(tree), 
			 "button-release-event",
			 G_CALLBACK(GtkUI_summaryview_button_release_cb),
			 sview);

  g_signal_connect(G_OBJECT(sview->search_entry), "changed",
		   G_CALLBACK(GtkUI_summaryview_quick_search_changed_cb),
		   sview);
  g_signal_connect(G_OBJECT(sview->search_entry), "key-press-event",
		   G_CALLBACK(GtkUI_summaryview_quick_search_key_press_cb),
		   sview);

  gtk_widget_show(tree);

  return sview;
}

static void
GtkUI_summaryview_free(XePGTKSummaryView *sview)
{
  if (!sview)
    return;

  gtk_widget_destroy(sview->widget);
  a_Folder_ref_free(sview->folder_ref);
  x_free(sview);
}

static void
GtkUI_summaryview_quick_search_show(XePGTKSummaryView *sview)
{
  gtk_widget_show(sview->search_box);
  gtk_widget_grab_focus(sview->search_entry);
}

static void
GtkUI_summaryview_quick_search_changed_cb(GtkEditable *editable,
					  gpointer asview)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) asview;
  const char *expr;

  /* Compile expression */
  expr = gtk_entry_get_text(GTK_ENTRY(sview->search_entry));
  if (sview->valid_preg)
    {
      regfree(&sview->preg);
      sview->valid_preg = FALSE;
    }

  if (expr && *expr)
    {
      sview->valid_preg = (0 == regcomp(&sview->preg, expr, 
					REG_ICASE | REG_NOSUB | REG_EXTENDED));
    }
  gtk_xtree_view_set_match_func(GTK_XTREE_VIEW(sview->tree),
				(expr && *expr ? 
				 GtkUI_summaryview_quick_search_func :
				 GtkUI_summaryview_match_new_func),
				sview);
  gtk_widget_set_sensitive(sview->search_icon, sview->valid_preg);
}

static gboolean
GtkUI_summaryview_quick_search_key_press_cb(GtkWidget *widget,
					    GdkEventKey *event,
					    gpointer asview)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) asview;
  
  switch (event->keyval)
    {
    case GDK_Escape:
      GtkUI_summaryview_quick_search_changed_cb(NULL, sview);
      gtk_widget_hide(sview->search_box);
      gtk_widget_grab_focus(sview->tree);
      return TRUE;
      break;
    case GDK_Return:
      GtkUI_summaryview_quick_search_changed_cb(NULL, sview);
      if ((event->state & GDK_MODIFIER_MASK) == 0)
	gtk_xtree_view_next_match(GTK_XTREE_VIEW(sview->tree), FALSE);
      else if (event->state & GDK_SHIFT_MASK)
	gtk_xtree_view_prev_match(GTK_XTREE_VIEW(sview->tree), FALSE);
      else if (event->state & GDK_CONTROL_MASK)
	gtk_xtree_view_activate_focused(GTK_XTREE_VIEW(sview->tree), 0);
      return TRUE;
    }
  return FALSE;
}

static gboolean
GtkUI_summaryview_quick_search_func(void *iterator, gpointer asview)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) asview;
  struct mail_envelopes_tree *msg_tree;

  msg_tree = (struct mail_envelopes_tree *) iterator;

  if (msg_tree && msg_tree->msg) 
    {
      char *dsp_subject = NULL;
      a_Message_message_get_subject(&msg_tree->msg->single_fields, 
				    GTK_DEST_CHARSET,
				    &dsp_subject);
      if (dsp_subject) 
	{
	  gboolean r;
	  r = (regexec(&sview->preg, dsp_subject, 0, NULL, 0) == 0);
	  free(dsp_subject);
	  return r;
	}
      return FALSE;
    }
  return FALSE;
}

static gboolean 
GtkUI_summaryview_match_new_func(void *iterator, gpointer asview)
{
  struct mail_envelopes_tree *msg_tree;

  msg_tree = (struct mail_envelopes_tree *) iterator;

  if (msg_tree && msg_tree->msg) 
    {
      if (!msg_tree->msg->flags || 
	  (msg_tree->msg->flags->flags & MAIL_FLAG_NEW) ||
	  !(msg_tree->msg->flags->flags & MAIL_FLAG_SEEN))
	return TRUE;
    }
  return FALSE;
}

static gboolean
GtkUI_summaryview_tree_data_func(void *iterator, 
				 guint col_id,
				 void **content,
				 gchar **attribute,
				 gpointer data)
{
  struct mail_envelopes_tree *msg_tree;
  mailmessage *msg = NULL;

  msg_tree =  (struct mail_envelopes_tree *) iterator;
  msg = msg_tree->msg;

  return GtkUI_summaryview_msg_data_func(msg, 
					 col_id, 
					 content, 
					 attribute, 
					 data);
}

static gboolean
GtkUI_summaryview_msg_data_func(mailmessage *msg, 
				guint col_id,
				void **content,
				gchar **attribute,
				gpointer data)
{
  static char buf[1024];
  char *dsp_subject = NULL;
  char *dsp_from = NULL;
  char *display_name = NULL;
  char *email_address = NULL;
  char *msg_date = NULL;
  const char *unit = "bkMGTPEZY?";
  size_t size;

  if (msg == NULL)
    {
      *content = NULL;
      *attribute = NULL;
      return TRUE;
    }

  switch (col_id)
    {
    case SV_STATUS_COL:
	  *content = (!msg || !msg->flags ? 
		      XEP_STOCK_NEW :
		      msg->flags->flags & MAIL_FLAG_DELETED ?
		      XEP_STOCK_DELETED :
		      msg->flags->flags & MAIL_FLAG_NEW ?
		      XEP_STOCK_NEW :
		      !(msg->flags->flags & MAIL_FLAG_SEEN) ?
		      XEP_STOCK_UNSEEN :
		      msg->flags->flags & MAIL_FLAG_ANSWERED ?
		      XEP_STOCK_REPLIED :
		      msg->flags->flags & MAIL_FLAG_FORWARDED ?
		      XEP_STOCK_FORWARDED :
		      NULL);
      break;
    case SV_FLAG_COL:
      *content =  ((msg && msg->flags && 
		    (msg->flags->flags & MAIL_FLAG_FLAGGED)) ?
		   XEP_STOCK_FLAG : NULL);
	break;
    case SV_SUBJECT_COL:
      a_Message_message_get_subject(&msg->single_fields, 
				    GTK_DEST_CHARSET,
				    &dsp_subject);
      if (dsp_subject)
	{
	  strncpy(buf, dsp_subject, sizeof(buf) - 1);
	  x_free(dsp_subject);
	  *content = buf;
	}
      else
	{
	  *content = NULL;
	}
      break;
    case SV_FROM_COL:
      a_Message_message_get_from(&msg->single_fields, GTK_DEST_CHARSET,
				 &display_name, &email_address);
      if (display_name)
	{
	  dsp_from = display_name;
	  x_free(email_address);
	}
      else if (email_address)
	{
	  dsp_from = email_address;
	}
      if (dsp_from)
	{
	  strncpy(buf, dsp_from, sizeof(buf) - 1);
	  x_free(dsp_from);
	  
	  *content = buf;
	}
      else
	{
	  *content = NULL;
	}
      break;
    case SV_DATE_COL:
      msg_date = 
	a_Message_message_formatted_get_local_date
	(&msg->single_fields, "%R %d-%m-%y", GTK_DEST_CHARSET);
      if (msg_date)
	{
	  strncpy(buf, msg_date, sizeof(buf) - 1);
	  x_free(msg_date);
	  *content = buf;
	}
      else
	{
	  *content = NULL;
	}

      break;
    case SV_SIZE_COL
:      /* b: byte
       * k: kilo byte = 1024 ~ 1e3
       * M: Mega byte = 1024*1024 ~ 1e6
       * G: Giga byte = 1024*1024*1024 ~ 1e9
       * P: Peta byte = 1024*1024*1024*1024 ~ 1e12
       * E: Exa byte  = 1024*1024*1024*1024*1024 ~ 1e15
       * Z: Zetta byte= 1024*1024*1024*1024*1024*1024 ~ 1e18
       * Y: Yotta byte= 1024*1024*1024*1024*1024*1024*1024 ~ 1e21 !
       */
      size = msg->size;
      for (; *unit && size > (1 << 10); unit++)
	{
	  size = size >> 10;
	}
      sprintf(buf, "%4d%c", size, *unit);
      *content = buf;
      break;

    }

  if (msg->flags && ((msg->flags->flags & MAIL_FLAG_NEW) ||
		     !(msg->flags->flags & MAIL_FLAG_SEEN)))
    {
      *attribute = "bold";
    }
  else
    {
      *attribute = NULL;
    }
  return TRUE;
}

static void
GtkUI_summaryview_tree_params_func(void *iterator, 
				   gboolean *has_children,
				   guint *depth,
				   gpointer data)
{
  struct mail_envelopes_tree *msg_tree;
  int i = 0;

  msg_tree = (struct mail_envelopes_tree *) iterator;

  *has_children = (msg_tree->children && carray_count(msg_tree->children) ? 
		   TRUE :
		   FALSE);
  while (msg_tree->parent)
    {
      msg_tree = msg_tree->parent;
      i++;
    }
  /* Do not count the root, i.e. root's children are at depth 0 */
  i--;
  *depth = i;
}

static gboolean
GtkUI_summaryview_tree_iterator_func(void **iterator, 
				     gboolean find_next,
				     gboolean sibling_first,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data)
{
  struct mail_envelopes_tree *msg_tree;
  gboolean found = FALSE;

  msg_tree = (struct mail_envelopes_tree *) *iterator;
  
  if (find_next)
    {
      do {
	msg_tree = mail_envelopes_tree_next(msg_tree, NULL, sibling_first);
	if (msg_tree)
	  found = TRUE; /* make it a function to filtered out items */
      } while (msg_tree && !found);
    }
  else
    {
      do {
	/* Do not ignore sibling_first to optimize tree traversal */
	msg_tree = mail_envelopes_tree_prev(msg_tree->parent, msg_tree);
	if (msg_tree)
	  found = TRUE;  /* make it a function to filtered out items */
      } while (msg_tree && !found);
    }

  if (found) 
    *iterator = msg_tree;
  return found;
}


static struct mail_envelopes_tree *
mail_envelopes_tree_next(struct mail_envelopes_tree *self,
			 struct mail_envelopes_tree *child,
			 int ignore_children)
{
  uint32_t len;

  if (self == NULL)
    return NULL;

  if (!ignore_children && 
      self->children && 
      (len = carray_count(self->children)))
    {
      carray *children = self->children;
      uint32_t i;
      
      if (child == NULL)
	return carray_get(children, 0);

      for (i = 0; i < len && carray_get(children, i) != child; i++)
	{
	  /* Empty on purpose */
	}
      if (i < len - 1)
	return carray_get(children, i + 1);
    }

  return mail_envelopes_tree_next(self->parent, self, 0);
}

static struct mail_envelopes_tree *
mail_envelopes_tree_prev(struct mail_envelopes_tree *parent,
			 struct mail_envelopes_tree *self)
{
  carray *siblings;
  struct mail_envelopes_tree *prev;
  uint32_t len;
  uint32_t i;
  
  if (parent == NULL)
    return NULL;

  siblings = parent->children;
  len = carray_count(siblings);

  for (i = 0; i < len && carray_get(siblings, i) != self; i++)
    {
      /* Empty on purpose */
    }
  if (i > 0)
    {
      /* find last descendent */
      prev = carray_get(siblings, i - 1);
      while (prev->children->len)
	{
	  prev = carray_get(prev->children, prev->children->len - 1);
	};
      return prev;
    }
  /* if parent is the root, we reached the top */
  if (parent && parent->parent)
    return parent;
  else
    return NULL;
}

#if 0
static gboolean
GtkUI_summaryview_list_data_func(void *iterator, 
				 guint col_id,
				 void **content,
				 gchar **attribute,
				 gpointer data)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) data;
  struct mail_envelopes_list *ep_list = sview->folder_ref->folder->ep_list;
  uint32_t index;
  mailmessage *msg = NULL;

  index = (uint32_t) iterator;
  if (index) 
    {
      /* iterator 0 is an invalid one */
      /* iterator start from 1 but array indexes must start from 0
       * so we adjust index to point into the array 
       */
      msg = carray_get(ep_list->tab, index - 1);
    }
  return GtkUI_summaryview_msg_data_func(msg,
					 col_id,
					 content, 
					 attribute, 
					 data);
}

static void
GtkUI_summaryview_list_params_func(void *iterator, 
				   gboolean *has_children,
				   guint *depth,
				   gpointer data)
{
  /* list item do not have any children and are at depth 0 */
  *has_children = FALSE;
  *depth = 0;

}

static gboolean
GtkUI_summaryview_list_iterator_func(void **iterator, 
				     gboolean find_next,
				     gboolean ignore_children,
				     guint sort_col_id,
				     GtkSortType sort_order,
				     gpointer data)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) data;
  struct mail_envelopes_list *ep_list = sview->folder_ref->folder->ep_list;
  gboolean found = FALSE;
  uint32_t index = (uint32_t) *iterator;

  if (find_next)
    {
      do
	{
	  index = mail_envelopes_list_next(ep_list, index);
	  if (index)
	    found = TRUE; /* change to a function to filter out items */
	  
	} while (index && !found);
    }
  else
    {
      do
	{
	  index = mail_envelopes_list_prev(ep_list, index);
	  if (index)
	    found = TRUE;
	} while (index && !found);
    }
  if (found)
    *iterator = (void *) index;

  return found;
}

static uint32_t
mail_envelopes_list_next(struct mail_envelopes_list *ep_list, uint32_t index)
{
  /* index starts from 1 but array index start from 0 */
  if (index < carray_count(ep_list->tab))
    return index + 1;
  else
    return 0;
}

static uint32_t
mail_envelopes_list_prev(struct mail_envelopes_list *ep_list, uint32_t index)
{
  /* index starts from 1 but array index start from 0 */
  if (index - 1 > 0)
    return index - 1;
  else
    return 0;
}
#endif

static mailmessage *
GtkUI_summaryview_get_selected_msg(XePGTKSummaryView *sview)
{
  struct mail_envelopes_tree *msg_tree;

  msg_tree = gtk_xtree_view_get_focused(GTK_XTREE_VIEW(sview->tree));

  if (msg_tree)
    return msg_tree->msg;
  else
    return NULL;
}

static void
GtkUI_summaryview_row_activated_cb(GtkXTreeView *xtv, void *iterator, guint col_id, gpointer asview)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) asview;
  XePGTKMessageView *mview  = (XePGTKMessageView *) sview->uimain->messageview;
  struct mail_envelopes_tree *msg_tree = (struct mail_envelopes_tree *) iterator;
  mailmessage *msg;
  XePMessage *message;

  if (!(msg_tree && msg_tree->msg)) 
    return;

  msg = msg_tree->msg;

  if (col_id == SV_FLAG_COL)
    {
      if (msg->flags)
	{
	  a_GtkUI_summaryview_selection_flag(sview->uimain, XEP_FLAGS_TOGGLE);
	}
      return;
    }
  message = a_Message_new_for_fetching(sview->folder_ref->folder, msg);
  if (message)
    {
      a_UI_messageview_show_message((XePUIMessageView *)mview, 
				    message);
    }
}

static gboolean
GtkUI_summaryview_button_release_cb(GtkWidget *widget,
				    GdkEventButton *event,
				    gpointer data)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) data;
  XePGTKMessageView *mview  = (XePGTKMessageView *) sview->uimain->messageview;
  XePUIMain *uimain = sview->uimain;
  int col_id;

  if (event->button == 2)
    {
      mailmessage *msg;
      XePMessage *message;
      msg = GtkUI_summaryview_get_selected_msg(sview);
      if (msg)
	{
	  if (gtk_xtree_view_find_cell(GTK_XTREE_VIEW(sview->tree),
				       event->x, event->y,
				       NULL,
				       NULL, &col_id, NULL))
	    {
	      if (col_id == SV_FLAG_COL)
		{
		  if (msg->flags)
		    {
		      /* Button 2 = unset */
		      a_GtkUI_summaryview_selection_flag(uimain, XEP_FLAGS_UNSET);
		    }
		  return TRUE;
		}
	    }
	  if (event->button == 2)
	    {
	      gchar *display;
	      display = a_GtkUtils_make_display_name(uimain->toplevel);
	      mview = 
		(XePGTKMessageView *) a_UI_messageview_create(uimain->xetpan,
							      display, 
							      uimain->type);
	      g_free(display);
	    }

	  message = a_Message_new_for_fetching(sview->folder_ref->folder, msg);
	  if (message)
	    {
	      a_UI_messageview_show_message((XePUIMessageView *)mview, 
					    message);
	    }

	}
      return TRUE;
    }
  return TRUE;
}

void
a_GtkUI_summaryview_set_view(XePGTKSummaryView *sview,
			     XePFolderRef *folder_ref)
{
  GtkUI_summaryview_set_view(sview, folder_ref, FALSE);
}

void
a_GtkUI_summaryview_set_view_async(XePGTKSummaryView *sview,
				   XePFolderRef *folder_ref)
{
  GtkUI_summaryview_set_view(sview, folder_ref, TRUE);
}

static void
GtkUI_summaryview_set_view(XePGTKSummaryView *sview,
			   XePFolderRef *folder_ref,
			   gboolean update_folder)
{
  gtk_widget_grab_focus(sview->tree);

  if (folder_ref)
    {
      gchar *toplevel_title;
      
      A_LOGGING_DEBUG("viewing summary of %s\n", folder_ref->ref);
      
      toplevel_title = g_strdup_printf(_("XetPan - %s"), folder_ref->ref);
      gtk_window_set_title(GTK_WINDOW(sview->uimain->toplevel),
			   toplevel_title);
      g_free(toplevel_title);
      x_free(sview->folder_ref->ref);
      sview->folder_ref->ref = X_STRDUP(folder_ref->ref);
      sview->folder_ref->folder = folder_ref->folder;
      gtk_xtree_view_freeze(GTK_XTREE_VIEW(sview->tree));
      if (update_folder)
	{
	  a_UI_update_folder(folder_ref->ref);
	}
      else
	{
	  GtkUI_summaryview_refresh(sview);
	}

    }
  else 
    {
      A_LOGGING_SYSTEM_ERROR("FolderRef == NULL");
      a_Folder_ref_free_content(sview->folder_ref);
    }
}

void
a_GtkUI_summaryview_selection_delete(XePUIMain *uimain, XePMessageFlagsOp flag_op)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  GtkUI_summaryview_flag_selection(sview, flag_op, MAIL_FLAG_DELETED);
}

void
a_GtkUI_summaryview_selection_flag(XePUIMain *uimain, XePMessageFlagsOp flag_op)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  GtkUI_summaryview_flag_selection(sview, flag_op, MAIL_FLAG_FLAGGED);
}

void
a_GtkUI_summaryview_selection_seen(XePUIMain *uimain, XePMessageFlagsOp flag_op)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  GtkUI_summaryview_flag_selection(sview, flag_op, MAIL_FLAG_SEEN);
}

static void
GtkUI_summaryview_flag_selection(XePGTKSummaryView *sview, 
				 XePMessageFlagsOp flag_op,
				 int flags)
{
  void **selection;
  uint32_t n;

  n = gtk_xtree_view_get_selection(GTK_XTREE_VIEW(sview->tree), &selection);

  if (n)
    { 
      carray *messages;
      gtk_xtree_view_freeze(GTK_XTREE_VIEW(sview->tree));
      
      messages = 
	a_Message_array_new_for_flags_change(sview->folder_ref->folder,
					     selection,
					     n, 
					     flag_op,
					     flags);

      free(selection);
      if (carray_count(messages))
	a_UI_folder_messages_change_flags(sview->folder_ref->folder, messages);
      else
	{
	  carray_free(messages);
	  gtk_xtree_view_thaw(GTK_XTREE_VIEW(sview->tree));
	}
    }
}

static void
GtkUI_summaryview_copy_done_cb(carray *folders, carray *messages, 
			       uint32_t failures, void *asview)
{
  XePMessage *message;
  uint32_t i;
  uint32_t count = carray_count(messages);

  A_LOGGING_TRACE;

  for (i = 0; i < count; i++)
    {
      message = (XePMessage *) carray_get(messages, i);
      a_Message_free(message);
    }
  carray_free(messages);
  
  carray_free(folders);

}

static gboolean
GtkUI_summaryview_move_selection_cb(void *asview, 
				    int response,
				    carray *folders)
{
  return TRUE;
}

static gboolean
GtkUI_summaryview_copy_selection_cb(void *asview, 
				    int response, 
				    carray *folders)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) asview;
  GtkWidget *toplevel = (GtkWidget *) sview->uimain->toplevel;
  void **selection;
  uint32_t n;

  if (response != GTK_RESPONSE_OK)
    {
      if (folders)
	carray_free(folders);
      return TRUE;
    }

  if (folders == NULL || carray_count(folders) == 0)
    {
      a_GtkUtils_dialog_alert(_("No Folders Selected"), toplevel,
			      XETPAN_WARNING_MESSAGE, 
			      _("Copy operation not executed."),
			      NULL,
			      XEP_GTKUTILS_OK, TRUE, TRUE, NULL, NULL);
      if (folders)
	carray_free(folders);
      return FALSE;
    }

  n = gtk_xtree_view_get_selection(GTK_XTREE_VIEW(sview->tree), &selection);

  if (n)
    { 
      carray *messages;

      messages = a_Message_array_new(sview->folder_ref->folder, selection, n);
      a_Folder_array_append_messages_async(folders, messages,
					   GtkUI_summaryview_copy_done_cb, 
					   sview);
      free(selection);
      return TRUE;
    }
  else
    {
      a_GtkUtils_dialog_alert(_("No Messages Selected"), toplevel,
			      XETPAN_WARNING_MESSAGE, 
			      _("Copy operation not executed."),
			      NULL,
			      XEP_GTKUTILS_OK, TRUE, TRUE, NULL, NULL);

      if (folders)
	carray_free(folders);
      return FALSE;
    }
}

static void
GtkUI_summaryview_copy_move_selection(XePGTKSummaryView *sview,
				      gboolean move)
{
  /* TODO: WARNING: beware of future standalone summary views */
  XePGtkUIMain *uimain = (XePGtkUIMain *) sview->uimain;
  GtkWidget **widget;

  widget = (move ? 
	    &uimain->move_folder_selector :
	    &uimain->copy_folder_selector);

  if (*widget == NULL)
    {
      XetPan *xetpan = (XetPan *) uimain->ui.xetpan;
      XePFolderSelectionCb user_cb;
      XePFoldersViewsList *foldersviewslist;
      GtkWidget *toplevel = (GtkWidget *) sview->uimain->toplevel;
      gchar *title;
      gchar *help = _("Select Destination Folders");
      
      if (move)
	{
	  user_cb = GtkUI_summaryview_move_selection_cb;
	  title = _("Move Messages");
	}
      else
	{
	  user_cb = GtkUI_summaryview_copy_selection_cb;
	  title = _("Copy Messages");
	}
	  
      foldersviewslist = xetpan->config->configuration->foldersviewslist;

      /* TODO: when gtkxtreeview implements storable state and selection
       * use only one widget, and instead store the two selections lists
       * and use the adequate selection list when showing the widget
       */
      *widget = a_GtkFoldersview_folder_selection(toplevel, title,  help,
						  foldersviewslist,
						  user_cb, sview);
    }
  else
    {
      gtk_widget_show(*widget);
    }
}

    

void
a_GtkUI_summaryview_expunge(XePUIMain *uimain)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;
  XePFolder *folder = sview->folder_ref->folder;

  gtk_xtree_view_freeze(GTK_XTREE_VIEW(sview->tree));
  a_UI_expunge_folder(folder->id);
}

void
a_GtkUI_summaryview_find_match(XePGTKSummaryView *sview, gboolean forward)
{
  if (forward)
    gtk_xtree_view_next_match(GTK_XTREE_VIEW(sview->tree), TRUE);
  else
    gtk_xtree_view_prev_match(GTK_XTREE_VIEW(sview->tree), TRUE);
}

void
a_GtkUI_folder_updated(XePUIMain *uimain, XePFolder *folder)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  A_LOGGING_DEBUG("folder '%s'\n", 
		  (folder ? a_Folder_get_id(folder) : "NULL"));

  if (sview->folder_ref->folder == folder || folder == NULL)
    {
      GtkUI_summaryview_refresh(sview);
    }
}

void 
a_GtkUI_folder_refresh(XePUIMain *uimain, XePFolder *folder)
{
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  A_LOGGING_DEBUG("folder '%s'\n", a_Folder_get_id(folder));

  if (sview->folder_ref->folder == folder || folder == NULL)
    {
      gtk_xtree_view_thaw(GTK_XTREE_VIEW(sview->tree));
    }
}

static void
GtkUI_summaryview_refresh(XePGTKSummaryView *sview)
{
  XePFolder *folder = sview->folder_ref->folder;

  a_GtkMessageview_message_clear((XePGTKMessageView*)sview->uimain->messageview);


  /* FIXME: We should try to redisplay the tree at the same position,
   * for this, there should be a mapping function from old tree to new tree
   * to transfer tree view status to it
   */
  /* TODO: fine grained locking management */
  gtk_xtree_view_reset(GTK_XTREE_VIEW(sview->tree));
  if (folder && folder->total && folder->ep_tree &&
      a_Folder_is_locked(folder) == 0)
    {
#ifdef VIEW_TREE
      gint length;
      /* We must count number of dummy messages */
      length = a_Folder_tree_get_total_count(folder);
      gtk_xtree_view_set_length(GTK_XTREE_VIEW(sview->tree), length);
      
      gtk_xtree_view_set_first_row(GTK_XTREE_VIEW(sview->tree), carray_get(folder->ep_tree->children, 0));
      
#else
      gtk_xtree_view_set_length(GTK_XTREE_VIEW(sview->tree), folder->total);
      gtk_xtree_view_set_first_row(GTK_XTREE_VIEW(sview->tree), (void *) 1);
#endif
    }
  else
    {
      gtk_xtree_view_set_length(GTK_XTREE_VIEW(sview->tree), 0);
    }
  
  gtk_xtree_view_thaw(GTK_XTREE_VIEW(sview->tree));
}

/* Message View */
XePUIMessageView *
a_GtkUI_messageview_create_standalone(XetPan *xetpan, 
				      char *terminal)
{
  return (XePUIMessageView *) a_GtkMessageview_create_standalone(xetpan, 
								 terminal, 
								 0);
}

void
a_GtkUI_messageview_close(XePUIMessageView *amview)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  A_LOGGING_TRACE;
  a_GtkMessageview_free(mview);
}

void
a_GtkUI_messageview_finalize(XePUIMessageView *amview, XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  a_GtkMessageview_finalize((XePGTKMessageView *) mview, message);

}

void
a_GtkUI_messageview_mime_view_show(XePUIMessageView *amview,
				   XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;
  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_view_show((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_view_block(XePUIMessageView *amview,
				    XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_view_block((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_view_unblock(XePUIMessageView *amview, 
				      XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_view_unblock((XePGTKMessageView *) mview, message);
}

void
a_GtkUI_messageview_mime_show_selected(XePUIMessageView *amview, 
				       XePMessage *message)
{
  XePGTKMessageView *mview = (XePGTKMessageView *) amview;

  if (mview->muid != message->muid)
    return;

  a_GtkMessageview_mime_show_selected((XePGTKMessageView *) mview, message);
}

/* Compose Window */
XePUIComposeWindow *
a_GtkUI_composewindow_create(XetPan *xetpan, char *terminal)
{
  return (XePUIComposeWindow *) a_GtkComposewindow_create(xetpan, terminal);
}

void
a_GtkUI_composewindow_set_message(XePUIComposeWindow *uicompose, 
				  XePMessage *message)
{
  a_GtkComposewindow_set_message((XePGTKComposeWindow *) uicompose, message);
}

void
a_GtkUI_composewindow_set_body_text(XePUIComposeWindow *uicompose,
				    const char *text,
				    size_t text_size,
				    const char *text_charset)
{
  a_GtkComposewindow_set_body_text((XePGTKComposeWindow *) uicompose, 
				   text,
				   text_size,
				   text_charset);
}

void
a_GtkUI_composewindow_add_header(XePUIComposeWindow *uicompose, 
				 XePMessageHeaderType header_type,
				 const char *header_name,
				 const char *header_value)
{
  a_GtkComposewindow_add_header((XePGTKComposeWindow *) uicompose,
				header_type,
				header_name,
				header_value);
}

void
a_GtkUI_composewindow_close(XePUIComposeWindow *uicompose)
{
  A_LOGGING_TRACE;
  a_GtkComposewindow_free((XePGTKComposeWindow *) uicompose);
}

/*************************************************************************/
/*                                                                       */
/* Watching IO traffic                                                   */
/*                                                                       */
/*************************************************************************/

int
a_GtkUI_create_io_watch(XePUI *ui, XePUIIOWatcher *watcher)
{
  GIOChannel *channel;
  GIOCondition gcondition = 0;
  short int condition = watcher->condition;
  unsigned int tag;

  A_LOGGING_TRACE;

  if (condition & XETPAN_POLLIN)
    gcondition |= G_IO_IN;
  if (condition & XETPAN_POLLOUT)
    gcondition |= G_IO_OUT;
  if (condition & XETPAN_POLLPRI)
    gcondition |= G_IO_PRI;
  if (condition & XETPAN_POLLERR)
    gcondition |= G_IO_ERR;
  if (condition & XETPAN_POLLHUP)
    gcondition |= G_IO_HUP;
  if (condition & XETPAN_POLLNVAL)
    gcondition |= G_IO_NVAL;
  
  channel = g_io_channel_unix_new(watcher->fd);
  
  g_io_channel_set_encoding(channel, NULL, NULL);
  
  tag = g_io_add_watch(channel, gcondition, GtkUI_io_watch_cb, watcher);
  
  watcher->channel = channel;
  watcher->tag = tag;
  
  return 0;
}

void
a_GtkUI_delete_io_watch(XePUI *ui, XePUIIOWatcher *watcher)
{
  A_LOGGING_DEBUG("deleting fd=%d", watcher->fd);

  g_source_remove(watcher->tag);
  g_io_channel_unref((GIOChannel *) watcher->channel);
}

static gboolean
GtkUI_io_watch_cb(GIOChannel *channel, GIOCondition gcondition, gpointer data)
{
  XePUIIOWatcher *watcher = (XePUIIOWatcher *) data;
  int condition = 0;

  if (gcondition & G_IO_IN)
    condition |= XETPAN_POLLIN;
  if (gcondition & G_IO_OUT)
    condition |= XETPAN_POLLOUT;
  if (gcondition & G_IO_PRI)
    condition |= XETPAN_POLLPRI;
  if (gcondition & G_IO_ERR)
    condition |= XETPAN_POLLERR;
  if (gcondition & G_IO_HUP)
    condition |= XETPAN_POLLHUP;
  if (gcondition & G_IO_NVAL)
    condition |= XETPAN_POLLNVAL;

  watcher->callback(watcher->fd, condition, watcher->data);
  return TRUE;
}

#endif /* USE_GTKUI */
