/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef UI_TYPES_H
#define UI_TYPES_H

#include <stdlib.h>
#include <cinthash.h>
#include <clist.h>
#include <carray.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

enum
  {
    XETPAN_EXEC_SYNC,
    XETPAN_EXEC_ASYNC
  };

enum
  {
    XETPAN_POLLIN   = 0x01,
    XETPAN_POLLPRI  = 0x02,
    XETPAN_POLLOUT  = 0x04,
    XETPAN_POLLERR  = 0x08,
    XETPAN_POLLHUP  = 0x10,
    XETPAN_POLLNVAL = 0x20
  };

typedef enum
  {
#ifdef USE_GTKUI
    XETPAN_UI_GTK,
#endif
#ifdef USE_TTYUI
    XETPAN_UI_TTY,
#endif
    XETPAN_UI_NB,
    XETPAN_UI_UNSPECIFIED,
    XETPAN_UI_QUIT
  } XePUIType;

typedef unsigned long XePUID;

typedef cinthash_t cinthash;

typedef struct XePUI_s XePUI;

struct XePUI_s
{
  XePUIType type;            /**< Default UI type            */
  int switch_state;          /**< Flag set when switching UI */
  char *terminal;
  cinthash *mainviews;       /**< List of main views         */
  cinthash *foldersviews;    /**< List of foldersviews       */
  clist *composewindows;     /**< List of compose windows    */
  cinthash *messageviews;    /**< List of message viewers    */
  carray *watched_fd;        /**< Watched file descriptors   */
  unsigned int max_wid;      /**< Maximum watch_id value     */
  int ideditor_lock;
  void *gtk;                 /**< A structure containing global GTK data */
  void *tty;                 /**< A structure containing global TTY data */
};

typedef struct XePUIIOWatcher_s XePUIIOWatcher;

struct XePUIIOWatcher_s
{
  unsigned int watch_id;
  int fd;
  void (*callback)(int afd, short int acondition, void *adata);
  void *data;
  void *channel;
  unsigned int tag;
  short int condition;
};

typedef struct XePUIMain_s XePUIMain;

/*! A Main View structure
 *  It is possible to have multiple main views (windows or consoles)
 */
struct XePUIMain_s
{
  XePUIType type;    /**< UI type                */
  XePUID uid;  /**< A unique identifier    */
  void *toplevel;    /**< The main view (window) */
  void *foldersview; /**< The foldersview        */
  void *summaryview; /**< The summary view       */
  void *messageview; /**< The message view       */
  void *menubar;     /**< The menu bar           */
  void *toolbar;     /**< The tool bar           */            
  void *statusbar;   /**< The status bar         */
  void *python;      /**< The Python console     */
  void *logview;     /**< The log viewer         */
  void *about;       /**< The about panel        */
  void *ideditor  ;  /**< The identities editor  */
  void *xetpan;      /**< The main XetPan struct */
  int kb_watch_id;   /**< The keyboard watch ID  */
  void *remote;      /**< The remote client      */
};

typedef struct XePUIFoldersView_s XePUIFoldersView;
struct XePUIFoldersView_s
{
  XePUIType type;    /**< UI type                */
  XePUID uid;
};

typedef union XePUIMessageView_u XePUIMessageView;

typedef union XePUIComposeWindow_u XePUIComposeWindow;

#endif
